"""Formula 6.2a from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM, MM2, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot2ShearResistance(Formula):
    r"""Class representing formula 6.2a for the calculation of the design value for the shear resistance, [$V_{Rd,c}$]."""

    label = "6.2ab"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        c_rd_c: DIMENSIONLESS,
        k: DIMENSIONLESS,
        rho_l: DIMENSIONLESS,
        f_ck: MPA,
        k_1: DIMENSIONLESS,
        sigma_cp: MPA,
        b_w: MM,
        d: MM,
        v_min: N,
    ) -> None:
        r"""[$V_{Rd,c}$] Design value for the shear resistance [$kN$].

        EN 1992-1-1:2004 art.6.2.2(1) - Formula (6.2a)

        Parameters
        ----------
        c_rd_c : DIMENSIONLESS
            [$C_{Rd,c}$] Coefficient for shear strength [$-$].
        k : DIMENSIONLESS
            [$k$] Size effect factor [$-$].
        rho_l : DIMENSIONLESS
            [$\rho_l$] Longitudinal reinforcement ratio [$-$].
        f_ck : MPA
            [$f_{ck}$] Characteristic compressive strength of concrete [$MPa$].
        k_1 : DIMENSIONLESS
            [$k_1$] Coefficient for concrete [$-$].
        sigma_cp : MPA
            [$\sigma_{cp}$] Compressive stress in the concrete [$MPa$].
        b_w : MM
            [$b_w$] Width of the web [$mm$].
        d : MM
            [$d$] Effective depth [$mm$].
        v_min : N
            [$v_{min}$] shear capacity without rebar [$N$].
        """
        super().__init__()
        self.c_rd_c = c_rd_c
        self.k = k
        self.rho_l = rho_l
        self.f_ck = f_ck
        self.k_1 = k_1
        self.sigma_cp = sigma_cp
        self.b_w = b_w
        self.d = d
        self.v_min = v_min

    @staticmethod
    def _evaluate(
        c_rd_c: DIMENSIONLESS,
        k: DIMENSIONLESS,
        rho_l: DIMENSIONLESS,
        f_ck: MPA,
        k_1: DIMENSIONLESS,
        sigma_cp: MPA,
        b_w: MM,
        d: MM,
        v_min: N,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            c_rd_c=c_rd_c,
            k=k,
            rho_l=rho_l,
            f_ck=f_ck,
            k_1=k_1,
            sigma_cp=sigma_cp,
            b_w=b_w,
            d=d,
            v_min=v_min,
        )
        result_62a = (c_rd_c * k * (100 * rho_l * f_ck) ** (1 / 3) + k_1 * sigma_cp) * b_w * d
        result_62b = (v_min + k_1 * sigma_cp) * b_w * d

        return max(result_62a, result_62b)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.2."""
        return LatexFormula(
            return_symbol=r"V_{Rd,c}",
            result=f"{self:.{n}f}",
            equation=r"\max(C_{Rd,c} \cdot k \cdot \left(100 \cdot \rho_l \cdot f_{ck}\right)^{1/3} + k_1 \cdot "
            r"\sigma_{cp}, v_{min} + k_1 \cdot \sigma_{cp}) \cdot b_w \cdot d",
            numeric_equation=rf"\max({self.c_rd_c:.{n}f} \cdot {self.k:.{n}f} \cdot \left(100 \cdot {self.rho_l:.{n}f} "
            rf"\cdot {self.f_ck:.{n}f}\right)^{{1/3}} + {self.k_1:.{n}f} \cdot {self.sigma_cp:.{n}f}, {self.v_min:.{n}f} + "
            rf"{self.k_1:.{n}f} \cdot {self.sigma_cp:.{n}f}) \cdot {self.b_w:.{n}f} \cdot {self.d:.{n}f}",
            comparison_operator_label="=",
            unit="N",
        )


class Form6Dot2aSub1ThicknessFactor(Formula):
    r"""Class representing formula 6.2a for k, the thickness factor, [$k$]."""

    label = "6.2aSub1"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        d: MM,
    ) -> None:
        r"""[$k$] factor to take thickness into account [$kN$].

        EN 1992-1-1:2004 art.6.2.2(1) - Formula (6.2a)

        Parameters
        ----------
        d : MM
            [$d$] Effective depth [$mm$].
        """
        super().__init__()
        self.d = d

    @staticmethod
    def _evaluate(
        d: MM,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(d=d)

        return min(1 + (200 / d) ** 0.5, 2.0)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.2a Sub 1."""
        return LatexFormula(
            return_symbol=r"k",
            result=f"{self:.{n}f}",
            equation=r"\min(1 + \sqrt{\frac{200}{d}}, 2.0)",
            numeric_equation=rf"\min(1 + \sqrt{{\frac{{200}}{{{self.d:.{n}f}}}}}, 2.0)",
            comparison_operator_label="=",
            unit="-",
        )


class Form6Dot2aSub2RebarRatio(Formula):
    r"""Class representing formula 6.2a for the tensile rebar ratio, [$\rho_l$]."""

    label = "6.2aSub2"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        a_sl: MM2,
        b_w: MM,
        d: MM,
    ) -> None:
        r"""[$\rho_l$] Tensile rebar ratio [$-$].

        EN 1992-1-1:2004 art.6.2.2(1) - Formula (6.2a)

        Parameters
        ----------
        a_sl : MM2
            [$A_{sl}$] Area of tensile reinforcement [$mm^2$].
        b_w : MM
            [$b_w$] Width of the web [$mm$].
        d : MM
            [$d$] Effective depth [$mm$].
        """
        super().__init__()
        self.A_sl = a_sl
        self.b_w = b_w
        self.d = d

    @staticmethod
    def _evaluate(
        a_sl: MM2,
        b_w: MM,
        d: MM,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a_sl=a_sl)
        raise_if_less_or_equal_to_zero(b_w=b_w, d=d)

        return min(a_sl / (b_w * d), 0.02)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.2a Sub 2."""
        return LatexFormula(
            return_symbol=r"\rho_l",
            result=f"{self:.{n}f}",
            equation=r"\min( \frac{A_{sl}}{b_w \cdot d}, 0.02)",
            numeric_equation=rf"\min( \frac{{{self.A_sl:.{n}f}}}{{{self.b_w:.{n}f} \cdot {self.d:.{n}f}}}, 0.02)",
            comparison_operator_label="=",
            unit="-",
        )



"""Testing formula 6.40 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_40 import Form6Dot40ReducedBendingMomentResistance
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot40ReducedBendingMomentResistance:
    """Validation for formula 6.40 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        mpl_z_rd = 6000.0
        n = 0.3
        a_f = 0.4

        formula = Form6Dot40ReducedBendingMomentResistance(mpl_z_rd=mpl_z_rd, n=n, a_f=a_f)
        manually_calculated_result = 5250.0  # Nmm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("mpl_z_rd", "n", "a_f"),
        [
            (-6000.0, 0.3, 0.4),  # mpl_z_rd is negative
            (6000.0, -0.3, 0.4),  # n is negative
            (6000.0, 0.3, -0.4),  # a_f is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, mpl_z_rd: float, n: float, a_f: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot40ReducedBendingMomentResistance(mpl_z_rd=mpl_z_rd, n=n, a_f=a_f)

    @pytest.mark.parametrize(
        ("mpl_z_rd", "n", "a_f"),
        [
            (6000.0, 0.3, 2.0),  # denominator (1 - 0.5 * a_f) <= 0
        ],
    )
    def test_raise_error_when_denominator_is_invalid(self, mpl_z_rd: float, n: float, a_f: float) -> None:
        """Test invalid denominator."""
        with pytest.raises(LessOrEqualToZeroError):
            Form6Dot40ReducedBendingMomentResistance(mpl_z_rd=mpl_z_rd, n=n, a_f=a_f)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"M_{N,z,Rd} = \min \left( M_{pl,z,Rd} \cdot \frac{1 - n}{1 - 0.5 \cdot a_f}, M_{pl,z,Rd} \right) = "
                r"\min \left( 6000.000 \cdot \frac{1 - 0.300}{1 - 0.5 \cdot 0.400}, 6000.000 \right) = 5250.000 \ Nmm",
            ),
            (
                "complete_with_units",
                r"M_{N,z,Rd} = \min \left( M_{pl,z,Rd} \cdot \frac{1 - n}{1 - 0.5 \cdot a_f}, M_{pl,z,Rd} \right) = "
                r"\min \left( 6000.000 \ Nmm \cdot \frac{1 - 0.300}{1 - 0.5 \cdot 0.400}, 6000.000 \ Nmm \right) = 5250.000 \ Nmm",
            ),
            ("short", r"M_{N,z,Rd} = 5250.000 \ Nmm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        mpl_z_rd = 6000.0
        n = 0.3
        a_f = 0.4

        latex = Form6Dot40ReducedBendingMomentResistance(mpl_z_rd=mpl_z_rd, n=n, a_f=a_f).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Tests for the CHS enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.chs import CHS


class TestCHS:
    """Tests for the CHS enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert CHS.CHS21_3x2_3.value == ("CHS 21.3x2.3", 21.3, 2.3)
        assert CHS.CHS2220x40.value == ("CHS 2220x40", 2220, 40)

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "CHS 21.3x2.3" in [e.value[0] for e in CHS]
        assert "CHS 2220x40" in [e.value[0] for e in CHS]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in CHS]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = CHS.CHS21_3x2_3
        assert profile.alias == "CHS 21.3x2.3"
        assert profile.diameter == 21.3
        assert profile.thickness == 2.3



"""Fixtures for testing steel cross sections."""

from collections.abc import Generator
from unittest.mock import MagicMock, patch

import pytest
from sectionproperties.post.post import SectionProperties

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_3_materials.table_3_1 import SteelStrengthClass
from blueprints.materials.steel import SteelMaterial
from blueprints.structural_sections.steel.steel_cross_sections._steel_cross_section import CombinedSteelCrossSection
from blueprints.structural_sections.steel.steel_cross_sections.chs_profile import CHSSteelProfile
from blueprints.structural_sections.steel.steel_cross_sections.i_profile import ISteelProfile
from blueprints.structural_sections.steel.steel_cross_sections.rhs_profile import RHSSteelProfile
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.chs import CHS
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.heb import HEB
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.rhs import RHS
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.strip import Strip
from blueprints.structural_sections.steel.steel_cross_sections.strip_profile import StripSteelProfile


@pytest.fixture
def mock_section_properties() -> Generator[MagicMock, None, None]:
    """Fixture to mock section properties with default values."""
    with patch(
        "blueprints.structural_sections.steel.steel_cross_sections.plotters.general_steel_plotter.CombinedSteelCrossSection.section_properties"
    ) as mock:
        mock.return_value = SectionProperties(ixx_c=1.0e6, iyy_c=1.0e6)
        yield mock


@pytest.fixture
def strip_profile() -> StripSteelProfile:
    """Fixture to set up a Strip profile for testing."""
    profile = Strip.STRIP160x5
    steel_class = SteelStrengthClass.S355
    return StripSteelProfile.from_standard_profile(profile=profile, steel_material=SteelMaterial(steel_class))


@pytest.fixture
def chs_profile() -> CHSSteelProfile:
    """Fixture to set up a CHS profile for testing."""
    profile: CHS = CHS.CHS508x16
    steel_class: SteelStrengthClass = SteelStrengthClass.S355
    return CHSSteelProfile.from_standard_profile(profile=profile, steel_material=SteelMaterial(steel_class))


@pytest.fixture
def i_profile() -> ISteelProfile:
    """Fixture to set up an I-profile for testing."""
    profile = HEB.HEB360
    steel_class = SteelStrengthClass.S355
    return ISteelProfile.from_standard_profile(profile=profile, steel_material=SteelMaterial(steel_class))


@pytest.fixture
def empty_combined_steel_cross_section() -> CombinedSteelCrossSection:
    """Fixture to set up a combined steel cross-section for testing."""
    return CombinedSteelCrossSection(
        name="Empty Combined Steel Cross Section",
    )


@pytest.fixture
def rhs_profile() -> RHSSteelProfile:
    """Fixture to set up an RHS profile for testing."""
    profile = RHS.RHS400x200_16
    steel_class = SteelStrengthClass.S355
    return RHSSteelProfile.from_standard_profile(profile=profile, steel_material=SteelMaterial(steel_class))



"""Tests for Table3Dot1NominalValuesHotRolledStructuralSteel class."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_3_materials.table_3_1 import (
    SteelStandardGroup,
    SteelStrengthClass,
    Table3Dot1NominalValuesHotRolledStructuralSteel,
)
from blueprints.type_alias import MM, MPA


class TestTable3Dot1NominalValuesHotRolledStructuralSteel:
    """Tests for the Table3Dot1NominalValuesHotRolledStructuralSteel class."""

    def test_valid_initialization(self) -> None:
        """Test that valid parameters create the instance successfully."""
        # Test with common steel type and thickness
        steel = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 30)
        assert steel.steel_class == SteelStrengthClass.S355
        assert steel.thickness == 30

    def test_invalid_steel_class(self) -> None:
        """Test that an invalid steel class raises ValueError."""

        # Creating a mock steel class that doesn't exist in the data
        class MockSteelClass:
            pass

        with pytest.raises(ValueError) as excinfo:
            Table3Dot1NominalValuesHotRolledStructuralSteel(MockSteelClass(), 30)  # type: ignore[arg-type]

        assert "Invalid steel class" in str(excinfo.value)

    def test_invalid_thickness_negative(self) -> None:
        """Test that negative thickness raises ValueError."""
        with pytest.raises(ValueError) as excinfo:
            Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, -5)

        assert "Thickness must be a positive number" in str(excinfo.value)

    def test_invalid_thickness_zero(self) -> None:
        """Test that zero thickness raises ValueError."""
        with pytest.raises(ValueError) as excinfo:
            Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 0)

        assert "Thickness must be a positive number" in str(excinfo.value)

    def test_invalid_thickness_type(self) -> None:
        """Test that non-numeric thickness raises ValueError."""
        with pytest.raises(ValueError) as excinfo:
            Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, "30 mm")  # type: ignore[arg-type]

        assert "Thickness must be a positive number" in str(excinfo.value)

    def test_fy_small_thickness(self) -> None:
        """Test yield strength for thickness ≤ 40 mm."""
        # Test standard steel type with thickness below 40 mm
        steel = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 40)
        assert steel.fy == 355

    def test_fy_large_thickness(self) -> None:
        """Test yield strength for thickness > 40 mm and ≤ 80 mm."""
        # Test standard steel type with thickness above 40 mm
        steel = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 50)
        assert steel.fy == 335

    def test_fy_boundary_thickness(self) -> None:
        """Test yield strength at the boundary (40 mm and 40.01mm)."""
        # Test at exactly 40 mm (should use the ≤40 mm values)
        steel_at_40 = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 40)
        assert steel_at_40.fy == 355

        # Test just above 40 mm (should use the >40 mm values)
        steel_above_40 = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 40.01)
        assert steel_above_40.fy == 335

    def test_fy_too_large_thickness(self) -> None:
        """Test yield strength for thickness > 80 mm raises error."""
        with pytest.raises(ValueError) as excinfo:
            _ = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 81).fy

        assert "exceeds maximum supported value of 80 mm" in str(excinfo.value)

    def test_fy_en10219_1_above_40mm(self) -> None:
        """Test that EN 10219-1 steels with thickness > 40 mm return ValueError for fy."""
        steel = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355_MH_MLH_10219_1, 50)
        with pytest.raises(ValueError):
            _ = steel.fy

    def test_fu_small_thickness(self) -> None:
        """Test ultimate tensile strength for thickness ≤ 40 mm."""
        # Test standard steel type with thickness below 40 mm
        steel = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 40)
        assert steel.fu == 490

    def test_fu_large_thickness(self) -> None:
        """Test ultimate tensile strength for thickness > 40 mm and ≤ 80 mm."""
        # Test standard steel type with thickness above 40 mm
        steel = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 50)
        assert steel.fu == 470

    def test_fu_too_large_thickness(self) -> None:
        """Test ultimate tensile strength for thickness > 80 mm raises error."""
        with pytest.raises(ValueError) as excinfo:
            _ = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 90).fu

        assert "exceeds maximum supported value of 80 mm" in str(excinfo.value)

    def test_fu_boundary_thickness(self) -> None:
        """Test ultimate tensile strength at the boundary (40 mm and 40.01mm)."""
        # Test at exactly 40 mm (should use the ≤40 mm values)
        steel_at_40 = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 40)
        assert steel_at_40.fu == 490

        # Test just above 40 mm (should use the >40 mm values)
        steel_above_40 = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 40.01)
        assert steel_above_40.fu == 470

    def test_fu_en10219_1_above_40mm(self) -> None:
        """Test that EN 10219-1 steels with thickness > 40 mm return ValueError for fu."""
        steel = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355_MH_MLH_10219_1, 50)

        with pytest.raises(ValueError):
            _ = steel.fu

    def test_steel_strength_class_display_name(self) -> None:
        """Test that steel strength classes have the correct display names."""
        assert SteelStrengthClass.S355.display_name == "S 355"
        assert SteelStrengthClass.S275_M_ML_10025_4.display_name == "S 275 M/ML"
        assert SteelStrengthClass.S460_NH_NLH_10219_1.display_name == "S 460 NH/NLH"

    def test_steel_strength_class_standard_group(self) -> None:
        """Test that steel strength classes belong to the correct standard group."""
        assert SteelStrengthClass.S355.standard_group == SteelStandardGroup.EN_10025_2
        assert SteelStrengthClass.S275_M_ML_10025_4.standard_group == SteelStandardGroup.EN_10025_4
        assert SteelStrengthClass.S355_MH_MLH_10219_1.standard_group == SteelStandardGroup.EN_10219_1

    @pytest.mark.parametrize(
        ("steel_class", "thickness", "expected_fy", "expected_fu"),
        [
            # EN 10025-2
            (SteelStrengthClass.S235, 30, 235, 360),
            (SteelStrengthClass.S235, 50, 215, 360),
            (SteelStrengthClass.S355, 35, 355, 490),
            (SteelStrengthClass.S355, 75, 335, 470),
            # EN 10025-3
            (SteelStrengthClass.S275_N_NL_10025_3, 20, 275, 390),
            (SteelStrengthClass.S275_N_NL_10025_3, 60, 255, 370),
            # EN 10025-4
            (SteelStrengthClass.S420_M_ML_10025_4, 10, 420, 520),
            (SteelStrengthClass.S420_M_ML_10025_4, 70, 390, 500),
            # EN 10025-5
            (SteelStrengthClass.S235_W_10025_5, 25, 235, 360),
            (SteelStrengthClass.S235_W_10025_5, 45, 215, 340),
            # EN 10025-6
            (SteelStrengthClass.S460_Q_QL_QL1_10025_6, 15, 460, 570),
            (SteelStrengthClass.S460_Q_QL_QL1_10025_6, 55, 440, 550),
            # EN 10210-1
            (SteelStrengthClass.S235_H_10210_1, 5, 235, 360),
            (SteelStrengthClass.S235_H_10210_1, 65, 215, 340),
        ],
    )
    def test_steel_strength_values(
        self,
        steel_class: SteelStrengthClass,
        thickness: MM,
        expected_fy: MPA,
        expected_fu: MPA,
    ) -> None:
        """Test that steel strength values match expected values for various classes and thicknesses."""
        steel = Table3Dot1NominalValuesHotRolledStructuralSteel(steel_class, thickness)
        assert steel.fy == expected_fy
        assert steel.fu == expected_fu

    def test_string_representation(self) -> None:
        """Test that __str__ returns the expected string format."""
        steel = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.S355, 30)
        expected_str = "S 355, t=30 mm, fy=355 N/mm², fu=490 N/mm²"
        assert str(steel) == expected_str



"""Formula 8.18 from EN 1992-1-1:2004: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM
from blueprints.validations import raise_if_negative


class Form8Dot18DesignValueTransmissionLength2(Formula):
    r"""Class representing formula 8.18 for the calculation of design value 2 of the transmission length [$l_{pt2}$]. The less favourable of
    [$l_{pt1}$] or [$l_{pt2}$] has to be chosen depending on the design situation.
    """

    label = "8.18"
    source_document = EN_1992_1_1_2004

    def __init__(self, l_pt: MM) -> None:
        r"""[$l_{pt2}$] design value 2 of the transmission length [$mm$].

        EN 1992-1-1:2004 art.8.10.2.2(3) - Formula (8.18)

        Parameters
        ----------
        l_pt : MM
            [$l_{pt}$] Basic value of the transmission length [$mm$].
            Use your own implementation for this value or use :class:`Form8Dot16BasicTransmissionLength` class.
        """
        super().__init__()
        self.l_pt = l_pt

    @staticmethod
    def _evaluate(
        l_pt: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(l_pt=l_pt)
        return 1.2 * l_pt

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 8.18."""
        return LatexFormula(
            return_symbol=r"l_{pt2}",
            result=f"{self:.{n}f}",
            equation=r"1.2 \cdot l_{pt}",
            numeric_equation=rf"1.2 \cdot {self.l_pt:.{n}f}",
            comparison_operator_label="=",
        )



"""Testing formula 5.38a of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_38a import Form5Dot38aCheckRelativeSlendernessRatio
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot38aCheckRelativeSlendernessRatio:
    """Validation for formula 5.38a from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        lambda_y = 1.0
        lambda_z = 1.5

        # Object to test
        formula = Form5Dot38aCheckRelativeSlendernessRatio(lambda_y=lambda_y, lambda_z=lambda_z)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("lambda_y", "lambda_z"),
        [
            (-1.0, 1.5),  # lambda_y is negative
            (1.0, -1.5),  # lambda_z is negative
            (0.0, 1.5),  # lambda_y is zero
            (1.0, 0.0),  # lambda_z is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, lambda_y: float, lambda_z: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form5Dot38aCheckRelativeSlendernessRatio(lambda_y=lambda_y, lambda_z=lambda_z)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \left( \frac{\lambda_{y}}{\lambda_{z}} \leq 2 \text{ and } \frac{\lambda_{z}}{\lambda_{y}} \leq 2 \right) \to "
                r"\left( \frac{1.000}{1.500} \leq 2 \text{ and } \frac{1.500}{1.000} \leq 2 \right) \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        lambda_y = 1.0
        lambda_z = 1.5

        # Object to test
        latex = Form5Dot38aCheckRelativeSlendernessRatio(lambda_y=lambda_y, lambda_z=lambda_z).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 3.7 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_7 import Form3Dot7NonLinearCreepCoefficient
from blueprints.validations import NegativeValueError


class TestForm3Dot7NonLinearCreepCoefficient:
    """Validation for formula 3.7 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        phi_inf_t0 = 0.25  # -
        k_sigma = 2.47  # days
        form_3_7 = Form3Dot7NonLinearCreepCoefficient(phi_inf_t0=phi_inf_t0, k_sigma=k_sigma)

        # Expected result, manually calculated
        manually_calculated_result = 5.174308

        assert form_3_7 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_phi_inf_t0_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        phi_inf_t0 = -0.25  # -
        k_sigma = 2.47  # days

        with pytest.raises(NegativeValueError):
            Form3Dot7NonLinearCreepCoefficient(phi_inf_t0=phi_inf_t0, k_sigma=k_sigma)

    def test_raise_error_when_negative_k_sigma_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        phi_inf_t0 = 0.25  # -
        k_sigma = -2.47  # days

        with pytest.raises(NegativeValueError):
            Form3Dot7NonLinearCreepCoefficient(phi_inf_t0=phi_inf_t0, k_sigma=k_sigma)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\varphi_{nl}(\infty, t_0) = \varphi(\infty, t_0) \cdot \exp( 1.5 ( k_{\sigma} - 0.45)) = 0.250 \cdot \exp( 1.5 ( 2.470 - 0.45)) = "
                r"5.174",
            ),
            ("short", r"\varphi_{nl}(\infty, t_0) = 5.174"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        phi_inf_t0 = 0.25  # -
        k_sigma = 2.47  # days

        # Object to test
        form_3_7_latex = Form3Dot7NonLinearCreepCoefficient(phi_inf_t0=phi_inf_t0, k_sigma=k_sigma).latex()

        actual = {"complete": form_3_7_latex.complete, "short": form_3_7_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 6.23 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MPA
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form6Dot23CheckShearStressInterface(Formula):
    r"""Class representing formula 6.23 for checking the shear stress at the interface between concrete cast."""

    label = "6.23"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        v_edi: MPA,
        v_rdi: MPA,
    ) -> None:
        r"""Check the shear stress at the interface between concrete cast.

        EN 1992-1-1:2004 art.6.2.5(1) - Formula (6.23)

        Parameters
        ----------
        v_edi : MPA
            [$v_{Edi}$] Design value of the shear stress at the interface [$MPa$].
        v_rdi : MPA
            [$v_{Rdi}$] Design shear strength of the interface [$MPa$].
        """
        super().__init__()
        self.v_edi = v_edi
        self.v_rdi = v_rdi

    @staticmethod
    def _evaluate(
        v_edi: MPA,
        v_rdi: MPA,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(v_edi=v_edi, v_rdi=v_rdi)

        return v_edi <= v_rdi

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.23."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=r"v_{Edi} \leq v_{Rdi}",
            numeric_equation=rf"{self.v_edi:.{n}f} \leq {self.v_rdi:.{n}f}",
            comparison_operator_label="\\to",
            unit="",
        )



"""Testing formula 9.5N of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_9_detailling_and_specific_rules.formula_9_5n import (
    Form9Dot5nMinimumShearReinforcementRatio,
)
from blueprints.validations import NegativeValueError


class TestForm9Dot5nMinimumShearReinforcementRatio:
    """Validation for formula 9.5N from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_ck = 30  # MPa
        f_yk = 500  # MPa
        form_9_5n = Form9Dot5nMinimumShearReinforcementRatio(f_ck=f_ck, f_yk=f_yk)

        # Expected result, manually calculated
        manually_calculated_result = 0.000876356

        assert form_9_5n == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_f_ck_is_given(self) -> None:
        """Test if error is raised when f_ck is negative."""
        f_ck = -30  # MPa
        f_yk = 500  # MPa

        with pytest.raises(NegativeValueError):
            Form9Dot5nMinimumShearReinforcementRatio(f_ck=f_ck, f_yk=f_yk)

    def test_raise_error_when_negative_f_yk_is_given(self) -> None:
        """Test if error is raised when f_yk is negative."""
        f_ck = 30  # MPa
        f_yk = -500  # MPa

        with pytest.raises(NegativeValueError):
            Form9Dot5nMinimumShearReinforcementRatio(f_ck=f_ck, f_yk=f_yk)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\rho_{w,min} = \left( 0.08 \cdot \sqrt{f_{ck}} \right) / f_{yk} = \left( 0.08 \cdot \sqrt{30.00} \right) / 500.00 = 0.000876",
            ),
            ("short", r"\rho_{w,min} = 0.000876"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ck = 30  # MPa
        f_yk = 500  # MPa

        # Object to test
        form_9_5n_latex = Form9Dot5nMinimumShearReinforcementRatio(f_ck=f_ck, f_yk=f_yk).latex()

        actual = {"complete": form_9_5n_latex.complete, "short": form_9_5n_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 7.4 from EN 1995-1-1:2004."""

from blueprints.codes.eurocode.en_1995_1_1_2004 import EN_1995_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, HZ, M_NS2
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form7Dot4VelocityResponseLimit(Formula):
    r"""Class representing formula 7.4 for the calculation of the upper limit of the velocity response of a unit impulse load."""

    label = "7.4"
    source_document = EN_1995_1_1_2004

    # Constants for validation
    b_50 = 50
    b_150 = 150

    def __init__(self, b: DIMENSIONLESS, f_1: HZ, ksi: DIMENSIONLESS) -> None:
        r"""[v_{limit}] Upper limit of the velocity response of a unit impulse load, in [$m/(Ns^2)$].

        EN 1995-1-1:2004 art 7.3.3(2) - Formula (7.4)

        Parameters
        ----------
        f_1 : HZ
            [$f_{1}$] Natural frequency of rectangular floor, laid freely on all four sides [$Hz$].
        b : DIMENSIONLESS
            [$b$] Dimensionless factor, taken as 120 in the Dutch National Annex [$-$].
        ksi : DIMENSIONLESS
            [$\xi$] Modal damping factor [$-$].

        Returns
        -------
        None
        """
        super().__init__()
        self.f_1 = f_1
        self.ksi = ksi
        self.b = b

    @staticmethod
    def _evaluate(b: DIMENSIONLESS, f_1: HZ, ksi: DIMENSIONLESS) -> M_NS2:
        """Evaluate the formula, for more information see the __init__ method."""
        if b < Form7Dot4VelocityResponseLimit.b_50 or b > Form7Dot4VelocityResponseLimit.b_150:
            raise ValueError(
                "b must be at least 50 and no less than 150 (For the Netherlands b must be exactly 120 according to Dutch National Annex)."
            )
        raise_if_less_or_equal_to_zero(f_1=f_1)
        raise_if_negative(ksi=ksi)
        return b ** (f_1 * ksi - 1)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.4."""
        eq_for: str = r"b^{(f_1 \cdot \xi - 1)}"
        repl_symb = {
            "b": f"{self.b:.{n}f}",
            "f_1": f"{self.f_1:.{n}f}",
            r"\xi": f"{self.ksi:.{n}f}",
        }
        return LatexFormula(
            return_symbol=r"v_{lim}",
            result=f"{self:.{n}f}",
            equation=eq_for,
            numeric_equation=latex_replace_symbols(eq_for, repl_symb),
        )



"""Testing formulas 6.37 and 6.38 from EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_37_38 import (
    Form6Dot37Dot38MomentReduction,
    Form6Dot38A,
    Form6Dot38N,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot37Dot38MomentReduction:
    """Validation for formulas 6.37 and 6.38 from EN 1993-1-1:2005."""

    def test_evaluation_37(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        mpl_z_rd = 60000.0
        a = 0.4
        n = 0.3

        # Object to test
        formula = Form6Dot37Dot38MomentReduction(mpl_z_rd=mpl_z_rd, a=a, n=n)

        # Expected result, manually calculated
        manually_calculated_result = 60000.0

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_38(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        mpl_z_rd = 60000.0
        a = 0.5
        n = 0.9666666667

        # Object to test
        formula = Form6Dot37Dot38MomentReduction(mpl_z_rd=mpl_z_rd, a=a, n=n)

        # Expected result, manually calculated
        manually_calculated_result = 7733.333333

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("mpl_z_rd", "n"),
        [
            (-60000.0, 0.5),  # mpl_z_rd is negative
            (60000.0, -0.5),  # n is negative
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, mpl_z_rd: float, n: float) -> None:
        """Test invalid negative values."""
        with pytest.raises(NegativeValueError):
            Form6Dot37Dot38MomentReduction(mpl_z_rd=mpl_z_rd, a=0.4, n=n)

    @pytest.mark.parametrize(
        "a",
        [
            0.0,  # a is zero
            -0.1,  # a is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float) -> None:
        """Test invalid zero or negative values."""
        with pytest.raises(LessOrEqualToZeroError):
            Form6Dot37Dot38MomentReduction(mpl_z_rd=60000.0, a=a, n=0.5)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"M_{N,z,Rd} = \begin{cases} M_{pl,z,Rd} & \text{if } n \leq a \\ "
                r"M_{pl,z,Rd} \cdot \left[1 - \left(\frac{ n - a}{1 - a}\right)^2\right] & \text{if } "
                r"n > a \end{cases} = "
                r"\begin{cases} 60000.000 & \text{if } 0.300 \leq 0.400 \\ "
                r"60000.000 \cdot \left[1 - \left(\frac{ 0.300 - 0.400}{1 - 0.400}\right)^2\right] & "
                r"\text{if } 0.300 > 0.400 \end{cases} = 60000.000 \ Nmm",
            ),
            (
                "complete_with_units",
                r"M_{N,z,Rd} = \begin{cases} M_{pl,z,Rd} & \text{if } n \leq a \\ "
                r"M_{pl,z,Rd} \cdot \left[1 - \left(\frac{ n - a}{1 - a}\right)^2\right] & \text{if } n > a \end{cases} = "
                r"\begin{cases} 60000.000 \ Nmm & \text{if } 0.300 \leq 0.400 \\ "
                r"60000.000 \ Nmm \cdot \left[1 - \left(\frac{ 0.300 - 0.400}{1 - 0.400}\right)^2\right] & "
                r"\text{if } 0.300 > 0.400 \end{cases} = 60000.000 \ Nmm",
            ),
            ("short", r"M_{N,z,Rd} = 60000.000 \ Nmm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        mpl_z_rd = 60000.0
        a = 0.4
        n = 0.3

        # Object to test
        latex = Form6Dot37Dot38MomentReduction(mpl_z_rd=mpl_z_rd, a=a, n=n).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot38N:
    """Validation for formula 6.38n from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        n_ed = 50000.0
        n_pl_rd = 100000.0

        # Object to test
        formula = Form6Dot38N(n_ed=n_ed, n_pl_rd=n_pl_rd)

        # Expected result, manually calculated
        manually_calculated_result = 0.5

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("n_ed", "n_pl_rd"),
        [
            (-50000.0, 100000.0),  # n_ed is negative
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, n_ed: float, n_pl_rd: float) -> None:
        """Test invalid negative values."""
        with pytest.raises(NegativeValueError):
            Form6Dot38N(n_ed=n_ed, n_pl_rd=n_pl_rd)

    @pytest.mark.parametrize(
        "n_pl_rd",
        [
            0.0,  # n_pl_rd is zero
            -0.1,  # n_pl_rd is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, n_pl_rd: float) -> None:
        """Test invalid zero or negative values."""
        with pytest.raises(LessOrEqualToZeroError):
            Form6Dot38N(n_ed=50000.0, n_pl_rd=n_pl_rd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"n = \frac{N_{Ed}}{N_{pl,Rd}} = \frac{50000.000}{100000.000} = 0.500 \ -",
            ),
            (
                "complete_with_units",
                r"n = \frac{N_{Ed}}{N_{pl,Rd}} = \frac{50000.000 \ N}{100000.000 \ N} = 0.500 \ -",
            ),
            ("short", r"n = 0.500 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        n_ed = 50000.0
        n_pl_rd = 100000.0

        # Object to test
        latex = Form6Dot38N(n_ed=n_ed, n_pl_rd=n_pl_rd).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot38A:
    """Validation for formula 6.38a from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        capital_a = 2000.0
        b = 100.0
        tf = 10.0

        # Object to test
        formula = Form6Dot38A(capital_a=capital_a, b=b, tf=tf)

        # Expected result, manually calculated
        manually_calculated_result = 0.00

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation2(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        capital_a = 20000.0
        b = 100.0
        tf = 10.0

        # Object to test
        formula = Form6Dot38A(capital_a=capital_a, b=b, tf=tf)

        # Expected result, manually calculated
        manually_calculated_result = 0.50

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("capital_a", "b", "tf"),
        [
            (0.0, 100.0, 10.0),  # capital_a is zero
            (-2000.0, 100.0, 10.0),  # capital_a is negative
            (2000.0, -100.0, 10.0),  # b is negative
            (2000.0, 100.0, -10.0),  # tf is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, capital_a: float, b: float, tf: float) -> None:
        """Test invalid zero or negative values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot38A(capital_a=capital_a, b=b, tf=tf)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"a = \min\left(\frac{A - 2 \cdot b \cdot t_f}{A}, 0.5\right) = \min\left(\frac{2000.000 - "
                r"2 \cdot 100.000 \cdot 10.000}{2000.000}, 0.5\right) = 0.000 \ -",
            ),
            (
                "complete_with_units",
                r"a = \min\left(\frac{A - 2 \cdot b \cdot t_f}{A}, 0.5\right) = \min\left(\frac{2000.000 \ mm^2 - "
                r"2 \cdot 100.000 \ mm \cdot 10.000 \ mm}{2000.000 \ mm^2}, 0.5\right) = 0.000 \ -",
            ),
            ("short", r"a = 0.000 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        capital_a = 2000.0
        b = 100.0
        tf = 10.0

        # Object to test
        latex = Form6Dot38A(capital_a=capital_a, b=b, tf=tf).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 8.8N from EN 1992-1-1:2004: Chapter 8 - Detailing of reinforcement and prestressing tendons."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, KN, MM, MPA
from blueprints.unit_conversion import N_TO_KN
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form8Dot8nAnchorageCapacityWeldedTransverseBar(Formula):
    """Class representing the formula 8.8N for the calculation of the anchorage capacity of welded transverse bar, welded on the inside of the main
    bar.
    """

    label = "8.8N"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        l_td: MM,
        diameter_t: MM,
        sigma_td: MPA,
        f_wd: KN,
    ) -> None:
        r"""[$F_{btd}$] Anchorage capacity of welded transverse bar, welded on the inside of the main bar [$kN$].

            Note: Value may be found in National Annex.

        EN 1992-1-1:2004 art.8.6(2) - formula (8.8N)

        Parameters
        ----------
        l_td: MM
            [$l_{td}$] Design length of transverse bar [$mm$].

            [$= 1.16 ⋅ ø_{t} ⋅ (f_{yd}/σ_{td})^{0.5} ≤ l_{t}$]

            Use your own implementation of this formula or use the SubForm8Dot8nDesignLengthOfTransverseBar class.
        diameter_t: MM
            [$ø_{t}$] Diameter of transverse bar [$mm$].
        sigma_td: MPA
            [$σ_{td}$] Concrete stress [$MPa$].

            [$=(f_{ctd}+σ_{cm})/y ≤ 3⋅f_{cd}$]

            Use your own implementation of this formula or use the SubForm8Dot8nConcreteStress class.
        f_wd: KN
            [$F_{wd}$] Design shear strength of weld (specified as a factor times [$A_{s}⋅f_{yd}$]; say [$0.5⋅A_{s}⋅f_{yd}$] where
            [$A_{s}$] is the cross-section of the anchored bar and fyd is its design yield strength)  [$kN$].
        """
        super().__init__()
        self.l_td = l_td
        self.diameter_t = diameter_t
        self.sigma_td = sigma_td
        self.f_wd = f_wd

    @staticmethod
    def _evaluate(
        l_td: MM,
        diameter_t: MM,
        sigma_td: MPA,
        f_wd: KN,
    ) -> KN:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(
            l_td=l_td,
            diameter_t=diameter_t,
            sigma_td=sigma_td,
            f_wd=f_wd,
        )
        return min(l_td * diameter_t * sigma_td * N_TO_KN, f_wd)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 8.8N."""
        return LatexFormula(
            return_symbol=r"F_{btd}",
            result=f"{self:.{n}f}",
            equation=r"\min\left( l_{td} \cdot Ø_t \cdot \sigma_{td}, F_{wd} \right)",
            numeric_equation=rf"\min\left( {self.l_td:.{n}f} \cdot {self.diameter_t:.{n}f} "
            rf"\cdot {self.sigma_td:.{n}f} / 1000, {self.f_wd:.{n}f} \right)",
            comparison_operator_label="=",
        )


class SubForm8Dot8nDesignLengthOfTransverseBar(Formula):
    """Class representing sub-formula for formula 8.8N, which calculates the design length of the transverse bar."""

    label = "8.8N"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        diameter_t: MM,
        f_yd: MPA,
        sigma_td: MPA,
        l_t: MM,
    ) -> None:
        r"""[$l_{td}$] Design length of transverse bar [$mm$].

        EN 1992-1-1:2004 art.8.6(2) - [$l_{td}$]

        Parameters
        ----------
        diameter_t: MM
            [$ø_{t}$] Diameter of transverse bar [$mm$].
        f_yd: MPA
            [$f_{yd}$] Design yield strength of bar [$MPa$].
        sigma_td: MPA
            [$σ_{td}$] Concrete stress [$MPa$].

            [$=(f_{ctd}+σ_{cm})/y ≤ 3⋅f_{cd}$]

            Use your own implementation of this formula or use the SubForm8Dot8nConcreteStress class.
        l_t: MM
            [$l_{t}$] Length of transverse bar, but not more than the spacing of bars to be anchored [$mm$].
        """
        super().__init__()
        self.diameter_t = diameter_t
        self.f_yd = f_yd
        self.sigma_td = sigma_td
        self.l_t = l_t

    @staticmethod
    def _evaluate(
        diameter_t: MM,
        f_yd: MPA,
        sigma_td: MPA,
        l_t: MM,
    ) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(
            diameter_t=diameter_t,
            f_yd=f_yd,
            l_t=l_t,
        )
        raise_if_less_or_equal_to_zero(sigma_td=sigma_td)
        return min(1.16 * diameter_t * (f_yd / sigma_td) ** 0.5, l_t)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 8.8N transverse bar."""
        return LatexFormula(
            return_symbol=r"l_{td}",
            result=f"{self:.{n}f}",
            equation=r"\min\left(l_t, 1.16 \cdot Ø_t \cdot ({\frac{f_{yd}}{\sigma_{td}}})^{0.5} \right)",
            numeric_equation=(
                rf"\min\left({self.l_t:.{n}f}, 1.16 \cdot {self.diameter_t:.{n}f} \cdot "
                rf"({{\frac{{{self.f_yd:.{n}f}}}{{{self.sigma_td:.{n}f}}}}})^{{0.5}} \right)"
            ),
            comparison_operator_label="=",
        )


class SubForm8Dot8nConcreteStress(Formula):
    """Class representing sub-formula for formula 8.8N, which calculates the concrete stress."""

    label = "8.8N"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_ctd: MPA,
        sigma_cm: MPA,
        y_function: DIMENSIONLESS,
        f_cd: MPA,
    ) -> None:
        r"""[$σ_{td}$] Concrete stress [$MPa$].

        EN 1992-1-1:2004 art.8.6(2) - [$σ_{td}$]

        Parameters
        ----------
        f_ctd: MPA
            [$f_{ctd}$] Design tensile strength of concrete [$MPa$].
        sigma_cm: MPA
            [$σ_{cm}$] Compression in the concrete perpendicular to both bars (mean value) [$MPa$].
        y_function: DIMENSIONLESS
            [$y$] A function [$-$]

            [$= 0.015 + 0.14 ⋅ exp(-0.18⋅x)$]

            Use your own implementation of this formula or use the SubForm8Dot8nFunctionY class.
        f_cd: MPA
            [$f_{cd}$] Design value compressive strength of concrete [$MPa$].
        """
        super().__init__()
        self.f_ctd = f_ctd
        self.sigma_cm = sigma_cm
        self.y_function = y_function
        self.f_cd = f_cd

    @staticmethod
    def _evaluate(
        f_ctd: MPA,
        sigma_cm: MPA,
        y_function: DIMENSIONLESS,
        f_cd: MPA,
    ) -> MPA:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(
            f_ctd=f_ctd,
            sigma_cm=sigma_cm,
            f_cd=f_cd,
        )
        raise_if_less_or_equal_to_zero(y_function=y_function)
        return min((f_ctd + sigma_cm) / y_function, 3 * f_cd)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 8.8N concrete stress."""
        return LatexFormula(
            return_symbol=r"\sigma_{td}",
            result=f"{self:.{n}f}",
            equation=r"\min\left( 3 \cdot f_{cd}, \frac{f_{ctd} + \sigma_{cm}}{y} \right)",
            numeric_equation=(
                rf"\min\left(3 \cdot {self.f_cd:.{n}f}, \frac{{{self.f_ctd:.{n}f} " rf"+ {self.sigma_cm:.{n}f}}}{{{self.y_function:.{n}f}}} \right)"
            ),
            comparison_operator_label="=",
        )


class SubForm8Dot8nFunctionY(Formula):
    """Class representing sub-formula for formula 8.8N, which calculates the function y."""

    label = "8.8N"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        x_function: DIMENSIONLESS,
    ) -> None:
        r"""[$y$] A function [$-$].

        EN 1992-1-1:2004 art.8.6(2) - [$y$]

        Parameters
        ----------
        x_function: DIMENSIONLESS
            [$x$] A function accounting for the geometry [$-$]

            [$= 2⋅(c/ø_{t}) + 1$]

            Use your own implementation of this formula or use the SubForm8Dot8nFunctionX class.
        """
        super().__init__()
        self.x_function = x_function

    @staticmethod
    def _evaluate(
        x_function: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(x_function=x_function)
        return 0.015 + 0.14 * np.exp(-0.18 * x_function)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 8.8N y-function."""
        return LatexFormula(
            return_symbol=r"y",
            result=f"{self:.{n}f}",
            equation=r"0.015 + 0.14 \cdot e^{-0.18 \cdot x}",
            numeric_equation=rf"0.015 + 0.14 \cdot e^{{-0.18 \cdot {self.x_function:.{n}f}}}",
            comparison_operator_label="=",
        )


class SubForm8Dot8nFunctionX(Formula):
    """Class representing sub-formula for formula 8.8N, which calculates the function x."""

    label = "8.8N"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        cover: MM,
        diameter_t: MM,
    ) -> None:
        r"""[$x$] A function accounting for the geometry [$-$].

        EN 1992-1-1:2004 art.8.6(2) - [$x$]

        Parameters
        ----------
        cover: MM
            [$c$] Concrete cover perpendicular to both bars [$mm$].
        diameter_t: MM
            [$ø_{t}$] Diameter of transverse bar [$mm$].
        """
        super().__init__()
        self.cover = cover
        self.diameter_t = diameter_t

    @staticmethod
    def _evaluate(
        cover: MM,
        diameter_t: MM,
    ) -> DIMENSIONLESS:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(cover=cover)
        raise_if_less_or_equal_to_zero(diameter_t=diameter_t)
        return 2 * (cover / diameter_t) + 1

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 8.8N x-function."""
        return LatexFormula(
            return_symbol=r"x",
            result=f"{self:.{n}f}",
            equation=r"2 \cdot \frac{c}{Ø_t}",
            numeric_equation=rf"2 \cdot \frac{{{self.cover:.{n}f}}}{{{self.diameter_t:.{n}f}}}",
            comparison_operator_label="=",
        )



"""Tests for the ABCEnumMeta class."""

from abc import ABCMeta, abstractmethod
from enum import Enum, EnumMeta

import pytest

from blueprints.utils.abc_enum_meta import ABCEnumMeta


class MockAbstractEnum(Enum, metaclass=ABCEnumMeta):
    """Mock class for testing the ABCEnumMeta class."""

    @abstractmethod
    def test_method(self) -> int:
        """Test method for the MockAbstractEnum class."""


class MockConcreteEnum(MockAbstractEnum):
    """Mock class for testing the ABCEnumMeta class."""

    A = "a"
    B = "b"

    def test_method(self) -> int:
        """Test method for the MockConcreteEnum class."""
        return 1


class TestABCEnumMeta:
    """Test class for ABCEnumMeta."""

    def test_inheritance(self) -> None:
        """Test if ABCEnumMeta inherits from ABCMeta and EnumMeta."""
        assert issubclass(ABCEnumMeta, ABCMeta)
        assert issubclass(ABCEnumMeta, EnumMeta)

    def test_mock_abstract_enum_is_abstract(self) -> None:
        """Test case for MockAbstractEnum class to check if it's an abstract class."""
        assert MockAbstractEnum.__abstractmethods__ == frozenset({"test_method"})

    def test_abstract_enum_with_memebers_raises_type_error(self) -> None:
        """Test case for MockAbstractEnumWithMembers class to check if it raises a TypeError."""
        with pytest.raises(TypeError) as error_info:

            class MockAbstractEnumWithMembers(Enum, metaclass=ABCEnumMeta):
                """Mock class with members for testing the ABCEnumMeta class."""

                A = "a"

                @abstractmethod
                def test_method(self) -> int:
                    """Test method for the MockAbstractEnum class."""

        assert (
            str(error_info.value)
            == "Can't instantiate abstract class 'MockAbstractEnumWithMembers' without an implementation for abstract method 'test_method'"
        )

    def test_mock_concrete_enum(self) -> None:
        """Test case for MockConcreteEnum class."""
        assert MockConcreteEnum.A.value == "a"
        assert MockConcreteEnum.B.value == "b"
        assert MockConcreteEnum.B.test_method() == 1

    def test_invalid_mock_concrete_enum(self) -> None:
        """Test case for InvalidMockConcreteEnum class."""
        with pytest.raises(TypeError) as error_info:

            class InvalidMockConcreteEnum(MockAbstractEnum):
                """Mock class for testing the ABCEnumMeta class."""

                A = "a"
                B = "b"

        assert str(error_info.value).endswith("without an implementation for abstract method 'test_method'")

    def test_abc_enum_meta_attribute_error(self) -> None:
        """Test instantiation of class with ABCEnumMeta where no abstract methods attribute is present."""

        class NoAbstractMethodsEnum(Enum, metaclass=ABCEnumMeta):
            ONE = 1
            TWO = 2

        assert NoAbstractMethodsEnum.ONE.value == 1



"""Testing formula 6.2 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_2 import (
    Form6Dot2aSub1ThicknessFactor,
    Form6Dot2aSub2RebarRatio,
    Form6Dot2ShearResistance,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot2ShearResistance:
    """Validation for formula 6.2a from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        c_rd_c = 0.18
        k = 1.0
        rho_l = 0.02
        f_ck = 30.0
        k_1 = 0.15
        sigma_cp = 1.0
        b_w = 300.0
        d = 500.0
        v_min = 0.035

        # Object to test
        formula = Form6Dot2ShearResistance(c_rd_c=c_rd_c, k=k, rho_l=rho_l, f_ck=f_ck, k_1=k_1, sigma_cp=sigma_cp, b_w=b_w, d=d, v_min=v_min)

        # Expected result, manually calculated
        manually_calculated_result = 128201.426312

        assert formula == pytest.approx(manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("c_rd_c", "k", "rho_l", "f_ck", "k_1", "sigma_cp", "b_w", "d", "v_min"),
        [
            (-0.18, 1.0, 0.02, 30.0, 0.15, 1.0, 300.0, 500.0, 0.035),  # c_rd_c is negative
            (0.18, -1.0, 0.02, 30.0, 0.15, 1.0, 300.0, 500.0, 0.035),  # k is negative
            (0.18, 1.0, -0.02, 30.0, 0.15, 1.0, 300.0, 500.0, 0.035),  # rho_l is negative
            (0.18, 1.0, 0.02, -30.0, 0.15, 1.0, 300.0, 500.0, 0.035),  # f_ck is negative
            (0.18, 1.0, 0.02, 30.0, -0.15, 1.0, 300.0, 500.0, 0.035),  # k_1 is negative
            (0.18, 1.0, 0.02, 30.0, 0.15, -1.0, 300.0, 500.0, 0.035),  # sigma_cp is negative
            (0.18, 1.0, 0.02, 30.0, 0.15, 1.0, -300.0, 500.0, 0.035),  # b_w is negative
            (0.18, 1.0, 0.02, 30.0, 0.15, 1.0, 300.0, -500.0, 0.035),  # d is negative
            (0.18, 1.0, 0.02, 30.0, 0.15, 1.0, 300.0, 500.0, -0.035),  # v_min is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(
        self, c_rd_c: float, k: float, rho_l: float, f_ck: float, k_1: float, sigma_cp: float, b_w: float, d: float, v_min: float
    ) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot2ShearResistance(c_rd_c=c_rd_c, k=k, rho_l=rho_l, f_ck=f_ck, k_1=k_1, sigma_cp=sigma_cp, b_w=b_w, d=d, v_min=v_min)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"V_{Rd,c} = \max(C_{Rd,c} \cdot k \cdot \left(100 \cdot \rho_l \cdot f_{ck}\right)^{1/3} + "
                r"k_1 \cdot \sigma_{cp}, v_{min} + k_1 \cdot \sigma_{cp}) \cdot b_w \cdot d"
                r" = \max(0.180 \cdot 1.000 \cdot \left(100 \cdot 0.020 \cdot 30.000\right)^{1/3} + 0.150 \cdot "
                r"1.000, 0.035 + 0.150 \cdot 1.000) \cdot 300.000 \cdot 500.000 = 128201.426 \ N",
            ),
            ("short", r"V_{Rd,c} = 128201.426 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        c_rd_c = 0.18
        k = 1.0
        rho_l = 0.02
        f_ck = 30.0
        k_1 = 0.15
        sigma_cp = 1.0
        b_w = 300.0
        d = 500.0
        v_min = 0.035

        # Object to test
        latex = Form6Dot2ShearResistance(c_rd_c=c_rd_c, k=k, rho_l=rho_l, f_ck=f_ck, k_1=k_1, sigma_cp=sigma_cp, b_w=b_w, d=d, v_min=v_min).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot2aSub1ThicknessFactor:
    """Validation for formula 6.2aSub1 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        d = 500.0

        # Object to test
        formula = Form6Dot2aSub1ThicknessFactor(d=d)

        # Expected result, manually calculated
        manually_calculated_result = 1.63245553203

        assert formula == pytest.approx(manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("d"),
        [
            (-500.0),  # d is negative
            (0.0),  # d is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, d: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot2aSub1ThicknessFactor(d=d)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"k = \min(1 + \sqrt{\frac{200}{d}}, 2.0) = \min(1 + \sqrt{\frac{200}{500.000}}, 2.0) = 1.632 \ -",
            ),
            ("short", r"k = 1.632 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        d = 500.0

        # Object to test
        latex = Form6Dot2aSub1ThicknessFactor(d=d).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot2aSub2RebarRatio:
    """Validation for formula 6.2aSub2 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a_sl = 1500.0
        b_w = 300.0
        d = 500.0

        # Object to test
        formula = Form6Dot2aSub2RebarRatio(a_sl=a_sl, b_w=b_w, d=d)

        # Expected result, manually calculated
        manually_calculated_result = 0.01

        assert formula == pytest.approx(manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a_sl", "b_w", "d"),
        [
            (-1500.0, 300.0, 500.0),  # A_sl is negative
            (1500.0, -300.0, 500.0),  # b_w is negative
            (1500.0, 0.0, 500.0),  # b_w is zero
            (1500.0, 300.0, -500.0),  # d is negative
            (1500.0, 300.0, 0.0),  # d is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a_sl: float, b_w: float, d: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot2aSub2RebarRatio(a_sl=a_sl, b_w=b_w, d=d)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\rho_l = \min( \frac{A_{sl}}{b_w \cdot d}, 0.02) = \min( \frac{1500.000}{300.000 \cdot 500.000}, 0.02) = 0.010 \ -",
            ),
            ("short", r"\rho_l = 0.010 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a_sl = 1500.0
        b_w = 300.0
        d = 500.0

        # Object to test
        latex = Form6Dot2aSub2RebarRatio(a_sl=a_sl, b_w=b_w, d=d).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Test formula 1.0.1 from NEN 9997-1+C2:2017: Chapter 1: General rules."""

import pytest

from blueprints.codes.eurocode.nen_9997_1_c2_2017.chapter_1_general_rules.formula_1_0_1 import Form1Dot0Dot1EquivalentPilePointCenterline
from blueprints.validations import LessOrEqualToZeroError


class TestForm1Dot0Dot1EquivalentPilePointCenterline:
    """Validation for formula 1.0.1 from NEN 9997-1+C2:2017."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        a = 0.3  # m
        b = 0.2  # m

        form_1_0_1 = Form1Dot0Dot1EquivalentPilePointCenterline(a=a, b=b)

        # manually calculated result
        manually_calculated_result = 0.2767923

        assert form_1_0_1 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "b"),
        [
            (-0.3, 0.45),
            (0.3, -0.45),
            (0.3, 0),
        ],
    )
    def test_raise_error_if_zero_or_negative(self, a: float, b: float) -> None:
        """Test if zero or negative values are given."""
        with pytest.raises(LessOrEqualToZeroError):
            Form1Dot0Dot1EquivalentPilePointCenterline(a=a, b=b)

    def test_raise_error_if_b_greater_than_1_5_times_a(self) -> None:
        """Test if b is greater than 1.5 times a."""
        with pytest.raises(ValueError):
            Form1Dot0Dot1EquivalentPilePointCenterline(a=0.3, b=0.5)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            (
                "complete",
                r"D_{eq} = 1.13 \cdot a \cdot \sqrt{\frac{min(b, 1.5 \cdot a)}{a}} = 1.13 \cdot 0.3 \cdot \sqrt\frac{0.2}{0.3} = 0.277",
            ),
            ("short", "D_{eq} = 0.277"),
            (
                "string",
                r"D_{eq} = 1.13 \cdot a \cdot \sqrt{\frac{min(b, 1.5 \cdot a)}{a}} = 1.13 \cdot 0.3 \cdot \sqrt\frac{0.2}{0.3} = 0.277",
            ),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a = 0.3  # m
        b = 0.2  # m

        # Object to test
        form = Form1Dot0Dot1EquivalentPilePointCenterline(a=a, b=b).latex()

        actual = {
            "complete": form.complete,
            "short": form.short,
            "string": str(form),
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



"""Rectangular cross-section shape."""

from dataclasses import dataclass

from sectionproperties.pre import Geometry
from shapely import Polygon

from blueprints.structural_sections._cross_section import CrossSection
from blueprints.type_alias import MM


@dataclass(frozen=True)
class RectangularCrossSection(CrossSection):
    """
    Class to represent a rectangular cross-section for geometric calculations.

    Parameters
    ----------
    width : MM
        The width of the rectangular cross-section.
    height : MM
        The height of the rectangular cross-section.
    x : MM
        The x-coordinate of the centroid of the rectangle. Default is 0.
    y : MM
        The y-coordinate of the centroid of the rectangle. Default is 0.
    name : str
        The name of the rectangular cross-section, default is "Rectangle".
    mesh_size : MM | None
        The maximum mesh size for the geometry. Default is 2.5 mm.
    """

    width: MM
    height: MM
    x: MM = 0
    y: MM = 0
    name: str = "Rectangle"

    def __post_init__(self) -> None:
        """Post-initialization to validate the width and height."""
        if self.width <= 0:
            raise ValueError(f"Width must be a positive value, but got {self.width}")
        if self.height <= 0:
            raise ValueError(f"Height must be a positive value, but got {self.height}")

    @property
    def polygon(self) -> Polygon:
        """
        Shapely Polygon representing the rectangular cross-section. Defines the coordinates of the rectangle based on width, height, x,
        and y. Counter-clockwise order.

        Returns
        -------
        Polygon
            The shapely Polygon representing the rectangle.
        """
        left_lower = (self.x - self.width / 2, self.y - self.height / 2)
        right_lower = (self.x + self.width / 2, self.y - self.height / 2)
        right_upper = (self.x + self.width / 2, self.y + self.height / 2)
        left_upper = (self.x - self.width / 2, self.y + self.height / 2)
        return Polygon(
            [
                left_lower,
                right_lower,
                right_upper,
                left_upper,
            ]
        )

    def geometry(
        self,
        mesh_size: MM | None = None,
    ) -> Geometry:
        """Return the geometry of the rectangular cross-section.

        Properties
        ----------
        mesh_size : MM
            Maximum mesh element area to be used within
            the Geometry-object finite-element mesh. If not provided, a default value will be used.

        """
        if mesh_size is None:
            minimum_mesh_size = 2.0
            mesh_length = max(min(self.width, self.height) / 20, minimum_mesh_size)
            mesh_size = mesh_length**2

        rectangular = Geometry(geom=self.polygon)
        rectangular.create_mesh(mesh_sizes=mesh_size)
        return rectangular



"""Testing formula 6.18 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_18 import Form6Dot18AdditionalTensileForce
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot18AdditionalTensileForce:
    """Validation for formula 6.18 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        v_ed = 100.0
        theta = 30.0
        alpha = 45.0

        # Object to test
        formula = Form6Dot18AdditionalTensileForce(v_ed=v_ed, theta=theta, alpha=alpha)

        # Expected result, manually calculated
        manually_calculated_result = 36.6025403784  # kN

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("v_ed", "theta", "alpha"),
        [
            (-100.0, 30.0, 45.0),  # v_ed is negative
            (100.0, -30.0, 45.0),  # theta is negative
            (100.0, 30.0, -45.0),  # alpha is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, v_ed: float, theta: float, alpha: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot18AdditionalTensileForce(v_ed=v_ed, theta=theta, alpha=alpha)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\Delta F_{td} = 0.5 \cdot V_{Ed} \cdot \left(\cot(\theta) - \cot(\alpha)\right) "
                r"= 0.5 \cdot 100.000 \cdot \left(\cot(30.000) - \cot(45.000)\right) = 36.603 \ kN",
            ),
            ("short", r"\Delta F_{td} = 36.603 \ kN"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        v_ed = 100.0
        theta = 30.0
        alpha = 45.0

        # Object to test
        latex = Form6Dot18AdditionalTensileForce(v_ed=v_ed, theta=theta, alpha=alpha).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Test the SteelElement class."""

from unittest.mock import Mock

import pytest

from blueprints.structural_sections.steel.steel_element import SteelElement


def test_name(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement name matches the mock cross-section name."""
    assert steel_element.cross_section.name == mock_cross_section.name


def test_area(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement area matches the mock cross-section area."""
    assert steel_element.cross_section.area == mock_cross_section.area


def test_perimeter(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement perimeter matches the mock cross-section perimeter."""
    assert steel_element.cross_section.perimeter == mock_cross_section.perimeter


def test_centroid(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement centroid matches the mock cross-section centroid."""
    assert steel_element.cross_section.centroid == mock_cross_section.centroid


def test_moment_of_inertia_about_y(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement moment of inertia about Y matches the mock cross-section."""
    assert steel_element.cross_section.moment_of_inertia_about_y == mock_cross_section.moment_of_inertia_about_y


def test_moment_of_inertia_about_z(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement moment of inertia about Z matches the mock cross-section."""
    assert steel_element.cross_section.moment_of_inertia_about_z == mock_cross_section.moment_of_inertia_about_z


def test_elastic_section_modulus_about_y_positive(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement elastic section modulus about Y+ matches the mock cross-section."""
    assert steel_element.cross_section.elastic_section_modulus_about_y_positive == mock_cross_section.elastic_section_modulus_about_y_positive


def test_elastic_section_modulus_about_y_negative(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement elastic section modulus about Y- matches the mock cross-section."""
    assert steel_element.cross_section.elastic_section_modulus_about_y_negative == mock_cross_section.elastic_section_modulus_about_y_negative


def test_elastic_section_modulus_about_z_positive(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement elastic section modulus about Z+ matches the mock cross-section."""
    assert steel_element.cross_section.elastic_section_modulus_about_z_positive == mock_cross_section.elastic_section_modulus_about_z_positive


def test_elastic_section_modulus_about_z_negative(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement elastic section modulus about Z- matches the mock cross-section."""
    assert steel_element.cross_section.elastic_section_modulus_about_z_negative == mock_cross_section.elastic_section_modulus_about_z_negative


def test_plastic_section_modulus_about_y(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement plastic section modulus about Y matches the mock cross-section."""
    assert steel_element.cross_section.plastic_section_modulus_about_y == mock_cross_section.plastic_section_modulus_about_y


def test_plastic_section_modulus_about_z(steel_element: SteelElement, mock_cross_section: Mock) -> None:
    """Test that the SteelElement plastic section modulus about Z matches the mock cross-section."""
    assert steel_element.cross_section.plastic_section_modulus_about_z == mock_cross_section.plastic_section_modulus_about_z


def test_weight_per_meter(steel_element: SteelElement, mock_cross_section: Mock, mock_material: Mock) -> None:
    """Test that the SteelElement weight per meter is calculated correctly."""
    expected_weight: float = mock_material.density * (mock_cross_section.area * 1e-6)
    assert steel_element.weight_per_meter == expected_weight


def test_yield_strength(steel_element: SteelElement, mock_material: Mock) -> None:
    """Test that the SteelElement yield strength matches the mock material yield strength."""
    assert steel_element.yield_strength == mock_material.yield_strength.return_value


def test_ultimate_strength(steel_element: SteelElement, mock_material: Mock) -> None:
    """Test that the SteelElement ultimate strength matches the mock material ultimate strength."""
    assert steel_element.ultimate_strength == mock_material.ultimate_strength.return_value


def test_invalid_yield_strength(steel_element: SteelElement, mock_material: Mock) -> None:
    """Test that accessing yield strength raises a ValueError if not defined."""
    mock_material.yield_strength.return_value = None
    with pytest.raises(ValueError, match="Yield strength is not defined for this material."):
        _ = steel_element.yield_strength


def test_invalid_ultimate_strength(steel_element: SteelElement, mock_material: Mock) -> None:
    """Test that accessing ultimate strength raises a ValueError if not defined."""
    mock_material.ultimate_strength.return_value = None
    with pytest.raises(ValueError, match="Ultimate strength is not defined for this material."):
        _ = steel_element.ultimate_strength



"""Formula 9.5N from EN 1992-1-1:2004: Chapter 9 - Detailing of members and particular rules."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form9Dot5nMinimumShearReinforcementRatio(Formula):
    """Class representing the formula 9.5N for the calculation of the minimum shear reinforcement ratio for beams."""

    label = "9.5N"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_ck: MPA,
        f_yk: MPA,
    ) -> None:
        r"""[$\rho_{w,min}$] Minimum shear reinforcement ratio for beams [$-$].

        EN 1992-1-1:2004 art.9.2.2(5) - Formula (9.5N)

        Parameters
        ----------
        f_ck: MPA
            [$f_{ck}$] Characteristic concrete compressive cylinder strength at 28 days [$MPa$].
        f_yk: MPA
            [$f_{yk}$] Characteristic yield strength reinforcement steel [$MPa$].
        """
        super().__init__()
        self.f_ck = f_ck
        self.f_yk = f_yk

    @staticmethod
    def _evaluate(f_ck: MPA, f_yk: MPA) -> DIMENSIONLESS:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(f_ck=f_ck, f_yk=f_yk)
        return (0.08 * np.sqrt(f_ck)) / f_yk

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 9.5N."""
        return LatexFormula(
            return_symbol=r"\rho_{w,min}",
            result=f"{self:.6f}",
            equation=r"\left( 0.08 \cdot \sqrt{f_{ck}} \right) / f_{yk}",
            numeric_equation=rf"\left( 0.08 \cdot \sqrt{{{self.f_ck:.{n}f}}} \right) / {self.f_yk:.{n}f}",
            comparison_operator_label="=",
        )



"""Formula 2.1a from NEN 9997-1+C2:2017: Chapter 2: Basis of geotechnical design."""

from blueprints.codes.eurocode.nen_9997_1_c2_2017 import NEN_9997_1_C2_2017
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_negative


class Form2Dot1aDesignValueLoad(Formula):
    """Class representing formula 2.1a for the calculation of the design value [$F_{d}$] of actions."""

    label = "2.1a"
    source_document = NEN_9997_1_C2_2017

    def __init__(self, gamma_f: DIMENSIONLESS, f_rep: DIMENSIONLESS) -> None:
        r"""[$F_{d}$] Design value of actions.

        NEN 9997-1+C2:2017 art.2.4.6.1(2) - (Formula 2.1a)

        Parameters
        ----------
        gamma_f : DIMENSIONLESS
            [$\gamma_{F}$] partial factor for actions for persistent and transient situations defined in annex A [$-$].
        f_rep : DIMENSIONLESS
            [$F_{rep}$] Representative value of actions.

            Use your own implementation for this value or use :class:`Form2Dot1bRepresentativeValue`.
        """
        super().__init__()
        self.gamma_f = gamma_f
        self.f_rep = f_rep

    @staticmethod
    def _evaluate(
        gamma_f: DIMENSIONLESS,
        f_rep: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(gamma_f=gamma_f)
        return gamma_f * f_rep

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 2.1a."""
        return LatexFormula(
            return_symbol=r"F_d",
            result=f"{self:.{n}f}",
            equation=r"\gamma_F \cdot F_{rep}",
            numeric_equation=rf"{self.gamma_f:.{n}f} \cdot {self.f_rep:.{n}f}",
            comparison_operator_label="=",
        )



r"""Formula 8.10 from EN 1992-1-1:2004: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_max_curly_brackets, latex_min_curly_brackets
from blueprints.type_alias import DIMENSIONLESS, MM
from blueprints.validations import raise_if_negative


class Form8Dot10DesignLapLength(Formula):
    r"""Class representing formula 8.10 for the calculation of the design lap length [$l_{0}$] [$mm$]."""

    label = "8.10"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        alpha_1: DIMENSIONLESS,
        alpha_2: DIMENSIONLESS,
        alpha_3: DIMENSIONLESS,
        alpha_5: DIMENSIONLESS,
        alpha_6: DIMENSIONLESS,
        l_b_rqd: MM,
        l_0_min: MM,
    ) -> None:
        r"""[$l_{0}$] Design lap length [$mm$].

        EN 1992-1-1:2004 art.8.7.3(1) - Formula (8.10)

        Parameters
        ----------
        alpha_1 : DIMENSIONLESS
            [$α_{1}$] Coefficient for the effect of the form of the bars assuming adequate cover (see figure 8.1) [$-$].

            [$= 1.0$] for bars in compression.

            [$= 1.0$] for straight bars in tension.

            [$= 1.0$] if [$c_{d} <= 3 ⋅ Ø$] for bars other than straight in tension (see figure 8.1 (b), (c) and (d)).

            [$= 0.7$] if [$c_{d} > 3 ⋅ Ø$] for bars other than straight in tension (see figure 8.1 (b), (c) and (d)).

            Note: see figure 8.3 for values of [$c_{d}$].
        alpha_2 : DIMENSIONLESS
            [$α_{2}$] Coefficient for the effect of minimum concrete cover (see figure 8.3) [$-$].

            [$= 1.0$] for bars in compression.

            [$= 1 - 0.15 ⋅ (c_{d} - Ø) / Ø <= 1$] with a minimum of 0.7 for straight bars in tension.

            [$= 1 - 0.15 ⋅ (c_{d} - 3 ⋅ Ø) / Ø <= 1$] with a minimum of 0.7 for bars other than straight in tension
            (see figure 8.1 (b), (c) and (d)).

            Note: see figure 8.3 for values of [$c_{d}$].
        alpha_3 : DIMENSIONLESS
            [$α_{3}$] Coefficient for the effect of confinement by transverse reinforcement [$-$].

            [$= 1.0$] for bars in compression.

            [$= 1 - K ⋅ λ <= 1$] with a minimum of 0.7 for bars in tension.

            Where: [$λ = (ΣA_{st} - ΣA_{st,min}) / A_{s}$].

            Where: [$ΣA_{st,min} = A_{s} ⋅ (σ_{sd}/f_{yd})$]

            With [$A_{s}$] = area of one lapped bar [$mm²$].

            Note: see figure 8.4 for values of [$K, A_{s}$] and [$A_{st}$].
        alpha_5 : DIMENSIONLESS
            [$α_{5}$] Coefficient for the effect of the pressure transverse to the plane of splitting along the design
            anchorage length [$l_{bd}$] (see 8.6) [$-$].

            [$= 1 - 0.04 ⋅ p <= 1$] with a minimum of 0.7 for all types of anchorage in compression.

            Where: p = transverse pressure at ultimate limit state along [$l_{bd}$] [$MPa$].
        alpha_6 : DIMENSIONLESS
            [$α_{6}$] Coefficient for the effect of reinforcement ratio [$-$].

            [$= (ρ_{1}/25)^{0.5} <= 1.5$] with a minimum of 1.0.

            Where: [$ρ_{1}$] = reinforcement percentage lapped within [$0,65 ⋅ l_{0}$] from the centre of the lap length
            considered (see figure 8.8) [$-$].

            Use your own implementation of this formula or use the :class:`SubForm8Dot10Alpha6` class.
        l_b_rqd : MM
            [$l_{b,rqd}$] Basic required anchorage length, for anchoring the force [$A_{s} ⋅ σ_{sd}$] in a straight bar assuming constant
            bond stress (formula 8.3) [$mm$].

            Use your own implementation for this value or use the :class:`Form8Dot3RequiredAnchorageLength` class.
        l_0_min : MM
            [$l_{0,min}$] Minimum design lap length [$mm$].

            [$= max(0.3 ⋅ α_{6} ⋅ l_{b,rqd}, 15 ⋅ Ø, 200)$] (formula 8.11).

            Use your own implementation of this formula or use :class:`Form8Dot11MinimumDesignLapLength` class.
        """
        super().__init__()
        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2
        self.alpha_3 = alpha_3
        self.alpha_5 = alpha_5
        self.alpha_6 = alpha_6
        self.l_b_rqd = l_b_rqd
        self.l_0_min = l_0_min

    @staticmethod
    def _evaluate(
        alpha_1: DIMENSIONLESS,
        alpha_2: DIMENSIONLESS,
        alpha_3: DIMENSIONLESS,
        alpha_5: DIMENSIONLESS,
        alpha_6: DIMENSIONLESS,
        l_b_rqd: MM,
        l_0_min: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_5=alpha_5,
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            l_0_min=l_0_min,
        )
        return max(alpha_1 * alpha_2 * alpha_3 * alpha_5 * alpha_6 * l_b_rqd, l_0_min)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns a LatexFormula representation of the formula."""
        return LatexFormula(
            return_symbol=r"l_{0}",
            result=f"{self:.{n}f}",
            equation=latex_max_curly_brackets(r"\alpha_1 \cdot \alpha_2 \cdot \alpha_3 \cdot \alpha_5 \cdot \alpha_6 \cdot l_{b,rqd}", r"l_{0,min}"),
            numeric_equation=latex_max_curly_brackets(
                rf"{self.alpha_1:.{n}f} \cdot {self.alpha_2:.{n}f} \cdot {self.alpha_3:.{n}f} \cdot "
                rf"{self.alpha_5:.{n}f} \cdot {self.alpha_6:.{n}f} \cdot {self.l_b_rqd:.{n}f}",
                f"{self.l_0_min:.{n}f}",
            ),
            comparison_operator_label="=",
        )


class SubForm8Dot10Alpha6(Formula):
    r"""Class representing the formula for the calculation of the coefficient [$α_{6}$]."""

    label = "8.8"
    source_document = EN_1992_1_1_2004

    def __init__(self, rho_1: DIMENSIONLESS) -> None:
        r"""[$α_{6}$] Coefficient for the effect of reinforcement ratio [$-$].

        EN 1992-1-1:2004 art.8.7.3(1) - Formula (8.8)

        Parameters
        ----------
        rho_1 : DIMENSIONLESS
            [$ρ_{1}$] Reinforcement percentage lapped within [$0,65 ⋅ l_{0}$] from the centre of the lap length
            considered (see figure 8.8) [$-$].
        """
        super().__init__()
        self.rho_1 = rho_1

    @staticmethod
    def _evaluate(rho_1: DIMENSIONLESS) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(rho_l=rho_1)
        value_max_1_5 = min((rho_1 / 25) ** 0.5, 1.5)
        return max(value_max_1_5, 1)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns a LatexFormula representation of the formula."""
        argument_1_formula = r"\left(\frac{\rho_1}{25}\right)^{0.5}"
        numerical_argument_1 = rf"\left(\frac{{{self.rho_1:.{n}f}}}{{25}}\right)^{{0.5}}"
        return LatexFormula(
            return_symbol=r"\alpha_6",
            result=f"{self:.{n}f}",
            equation=f"{latex_max_curly_brackets(latex_min_curly_brackets(argument_1_formula, '1.5'), '1')}",
            numeric_equation=f"{latex_max_curly_brackets(latex_min_curly_brackets(numerical_argument_1, '1.5'), '1')}",
            comparison_operator_label="=",
        )



"""Testing formula 7.9 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_9 import Form7Dot9EpsilonSmMinusEpsilonCm
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot9EpsilonSmMinusEpsilonCm:
    """Validation for formula 7.9 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        sigma_s = 400.0
        k_t = 0.6
        f_ct_eff = 2.5
        rho_p_eff = 0.02
        e_s = 200000.0
        e_cm = 30000.0

        # Object to test
        formula = Form7Dot9EpsilonSmMinusEpsilonCm(
            sigma_s=sigma_s,
            k_t=k_t,
            f_ct_eff=f_ct_eff,
            rho_p_eff=rho_p_eff,
            e_s=e_s,
            e_cm=e_cm,
        )

        # Expected result, manually calculated
        manually_calculated_result = 0.001575  # dimensionless

        assert formula == pytest.approx(manually_calculated_result, rel=1e-3)

    @pytest.mark.parametrize(
        ("sigma_s", "k_t", "f_ct_eff", "rho_p_eff", "e_s", "e_cm"),
        [
            (-400.0, 0.6, 2.5, 0.02, 200000.0, 30000.0),  # sigma_s is negative
            (400.0, -0.6, 2.5, 0.02, 200000.0, 30000.0),  # k_t is negative
            (400.0, 0.6, -2.5, 0.02, 200000.0, 30000.0),  # f_ct_eff is negative
            (400.0, 0.6, 2.5, -0.02, 200000.0, 30000.0),  # rho_p_eff is negative
            (400.0, 0.6, 2.5, 0.02, -200000.0, 30000.0),  # e_s is negative
            (400.0, 0.6, 2.5, 0.02, 200000.0, -30000.0),  # e_cm is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(
        self, sigma_s: float, k_t: float, f_ct_eff: float, rho_p_eff: float, e_s: float, e_cm: float
    ) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form7Dot9EpsilonSmMinusEpsilonCm(
                sigma_s=sigma_s,
                k_t=k_t,
                f_ct_eff=f_ct_eff,
                rho_p_eff=rho_p_eff,
                e_s=e_s,
                e_cm=e_cm,
            )

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\epsilon_{sm} - \epsilon_{cm} = \max\left(\frac{\sigma_s - k_t \cdot \frac{f_{ct,eff}}{\rho_{p,eff}} "
                r"\cdot \left(1 + \frac{E_s}{E_{cm}} \cdot \rho_{p,eff}\right)}{E_s}; \frac{0.6 \cdot \sigma_s}{E_s}\right)"
                r" = \max\left(\frac{400.000 - 0.600 \cdot \frac{2.500}{0.020} \cdot \left(1 + \frac{200000.000}{30000.000} "
                r"\cdot 0.020\right)}{200000.000}; \frac{0.6 \cdot 400.000}{200000.000}\right) = 0.001575 \ -",
            ),
            ("short", r"\epsilon_{sm} - \epsilon_{cm} = 0.001575 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        sigma_s = 400.0
        k_t = 0.6
        f_ct_eff = 2.5
        rho_p_eff = 0.02
        e_s = 200000.0
        e_cm = 30000.0

        # Object to test
        latex = Form7Dot9EpsilonSmMinusEpsilonCm(
            sigma_s=sigma_s,
            k_t=k_t,
            f_ct_eff=f_ct_eff,
            rho_p_eff=rho_p_eff,
            e_s=e_s,
            e_cm=e_cm,
        ).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 7.20 from EN 1992-1-1:2004: Chapter 7 - Serviceability Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form7Dot20EffectiveModulusCreep(Formula):
    r"""Class representing formula 7.20 for the calculation of [$E_{c,eff}$]."""

    label = "7.20"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        e_cm: MPA,
        phi_inf_t0: DIMENSIONLESS,
    ) -> None:
        r"""[$E_{c,eff}$] Calculation of the effective modulus of elasticity for concrete including creep.

        EN 1992-1-1:2004 art.7.4.3(5) - Formula (7.20)

        Parameters
        ----------
        e_cm : MPA
            [$E_{cm}$] Secant modulus of elasticity of concrete [$MPa$].
        phi_inf_t0 : DIMENSIONLESS
            [$\phi(\infty, t_0)$] Creep coefficient relevant for the load and time interval (see 3.1.3) [$-$].
        """
        super().__init__()
        self.e_cm = e_cm
        self.phi_inf_t0 = phi_inf_t0

    @staticmethod
    def _evaluate(
        e_cm: MPA,
        phi_inf_t0: DIMENSIONLESS,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(e_cm=e_cm, phi_inf_t0=phi_inf_t0)

        return e_cm / (1 + phi_inf_t0)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.20."""
        _equation: str = r"\frac{E_{cm}}{1 + \phi(\infty , t_0)}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"E_{cm}": f"{self.e_cm:.{n}f}",
                r"\phi(\infty , t_0)": f"{self.phi_inf_t0:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"E_{c,eff}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="MPa",
        )



"""Formula 6.30 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DEG, DIMENSIONLESS, MM, MM2, MPA, NMM
from blueprints.validations import raise_if_negative


class Form6Dot30DesignTorsionalResistanceMoment(Formula):
    r"""Class representing formula 6.30 for the calculation of the design torsional resistance moment, [$T_{Rd,max}$]."""

    label = "6.30"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        nu: DIMENSIONLESS,
        alpha_cw: DIMENSIONLESS,
        f_cd: MPA,
        a_k: MM2,
        t_ef_i: MM,
        theta: DEG,
    ) -> None:
        r"""[$T_{Rd,max}$] Design torsional resistance moment [$Nmm$].

        EN 1992-1-1:2004 art.6.2.3(4) - Formula (6.30)

        Parameters
        ----------
        nu : DIMENSIONLESS
            [$\nu$] Strength reduction factor for concrete cracked in shear, see 6.2.2 (6) [$-$].
        alpha_cw : DIMENSIONLESS
            [$\alpha_{cw}$] Coefficient taking account of the state of the stress in the compression chord, see Expression (6.9) [$-$].
        f_cd : MPA
            [$f_{cd}$] Design value of concrete compressive strength [$MPa$].
        a_k : MM2
            [$A_{k}$] Area enclosed by the centre-lines of the connecting walls, including inner hollow areas [$mm^2$].
        t_ef_i : MM
            [$t_{ef,i}$] Effective wall thickness [$mm$].
        theta : DEG
            [$\theta$] Angle of compression struts (see Figure 6.5) [$degrees$].
        """
        super().__init__()
        self.nu = nu
        self.alpha_cw = alpha_cw
        self.f_cd = f_cd
        self.a_k = a_k
        self.t_ef_i = t_ef_i
        self.theta = theta

    @staticmethod
    def _evaluate(
        nu: DIMENSIONLESS,
        alpha_cw: DIMENSIONLESS,
        f_cd: MPA,
        a_k: MM2,
        t_ef_i: MM,
        theta: DEG,
    ) -> NMM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            nu=nu,
            alpha_cw=alpha_cw,
            f_cd=f_cd,
            a_k=a_k,
            t_ef_i=t_ef_i,
            theta=theta,
        )

        return 2 * nu * alpha_cw * f_cd * a_k * t_ef_i * np.sin(np.deg2rad(theta)) * np.cos(np.deg2rad(theta))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.30."""
        return LatexFormula(
            return_symbol=r"T_{Rd,max}",
            result=f"{self:.{n}f}",
            equation=r"2 \cdot \nu \cdot \alpha_{cw} \cdot f_{cd} \cdot A_{k} \cdot t_{ef,i} \cdot \sin(\theta) \cdot \cos(\theta)",
            numeric_equation=rf"2 \cdot {self.nu:.{n}f} \cdot {self.alpha_cw:.{n}f} \cdot {self.f_cd:.{n}f} \cdot {self.a_k:.{n}f} "
            rf"\cdot {self.t_ef_i:.{n}f} \cdot \sin({self.theta:.{n}f}) \cdot \cos({self.theta:.{n}f})",
            comparison_operator_label="=",
            unit="Nmm",
        )



"""Formula 5.32 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KNM


class Form5Dot32EquivalentFirstOrderEndMoment(Formula):
    """Class representing formula 5.32 for the calculation of the equivalent first order end moment, [$M_{0e}$]."""

    label = "5.32"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        m_01: KNM,
        m_02: KNM,
    ) -> None:
        r"""[$M_{0e}$] Equivalent first order end moment [$kNm$].

        EN 1992-1-1:2004 art.5.8.8.2(2) - Formula (5.32)

        Parameters
        ----------
        m_01 : KNM
            [$M_{01}$] The smaller first order end moment [$kNm$].
        m_02 : KNM
            [$M_{02}$] The larger first order end moment [$kNm$].
        """
        super().__init__()
        self.m_01 = m_01
        self.m_02 = m_02

    @staticmethod
    def _evaluate(
        m_01: KNM,
        m_02: KNM,
    ) -> KNM:
        """Evaluates the formula, for more information see the __init__ method."""
        if abs(m_02) < abs(m_01):
            raise ValueError("The absolute value of M_02 must be equal to or larger than the absolute value of M_01.")

        return max(0.6 * m_02 + 0.4 * m_01, 0.4 * m_02)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.32."""
        return LatexFormula(
            return_symbol=r"M_{0e}",
            result=f"{self:.{n}f}",
            equation=r"\max\left(0.6 \cdot M_{02} + 0.4 \cdot M_{01}; 0.4 \cdot M_{02}\right)",
            numeric_equation=rf"\max\left(0.6 \cdot {self.m_02:.{n}f} + 0.4 \cdot {self.m_01:.{n}f}; 0.4 \cdot {self.m_02:.{n}f}\right)",
            comparison_operator_label="=",
            unit="kNm",
        )



"""Testing formula 6.30 of NEN-EN 1993-1-1+A1:2016."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_30 import Form6Dot30ReducedPlasticResistanceMoment
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot30ReducedPlasticResistanceMoment:
    """Validation for formula 6.30 from NEN-EN 1993-1-1+A1:2016."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        w_pl_y = 500000.0  # mm^3
        rho = 0.5  # dimensionless
        h_w = 300.0  # mm
        t_w = 10.0  # mm
        f_y = 355.0  # MPa
        gamma_m0 = 1.0  # dimensionless
        m_y_c_rd = 1e10  # Nmm

        # Object to test
        formula = Form6Dot30ReducedPlasticResistanceMoment(
            w_pl_y=w_pl_y,
            rho=rho,
            h_w=h_w,
            t_w=t_w,
            f_y=f_y,
            gamma_m0=gamma_m0,
            m_y_c_rd=m_y_c_rd,
        )

        # Expected result, manually calculated
        manually_calculated_result = 137562500.0  # Nmm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation2(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        w_pl_y = 500000.0  # mm^3
        rho = 0.5  # dimensionless
        h_w = 300.0  # mm
        t_w = 10.0  # mm
        f_y = 355.0  # MPa
        gamma_m0 = 1.0  # dimensionless
        m_y_c_rd = 100.0  # Nmm

        # Object to test
        formula = Form6Dot30ReducedPlasticResistanceMoment(
            w_pl_y=w_pl_y,
            rho=rho,
            h_w=h_w,
            t_w=t_w,
            f_y=f_y,
            gamma_m0=gamma_m0,
            m_y_c_rd=m_y_c_rd,
        )

        # Expected result, manually calculated
        manually_calculated_result = 100.0  # Nmm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("w_pl_y", "rho", "h_w", "t_w", "f_y", "gamma_m0", "m_y_c_rd"),
        [
            (-500000.0, 0.5, 300.0, 10.0, 355.0, 1.0, 100000000.0),  # w_pl_y is negative
            (500000.0, -0.5, 300.0, 10.0, 355.0, 1.0, 100000000.0),  # rho is negative
            (500000.0, 0.5, -300.0, 10.0, 355.0, 1.0, 100000000.0),  # h_w is negative
            (500000.0, 0.5, 300.0, -10.0, 355.0, 1.0, 100000000.0),  # t_w is negative
            (500000.0, 0.5, 300.0, 10.0, -355.0, 1.0, 100000000.0),  # f_y is negative
            (500000.0, 0.5, 300.0, 10.0, 355.0, -1.0, 100000000.0),  # gamma_m0 is negative
            (500000.0, 0.5, 300.0, 10.0, 355.0, 1.0, -100000000.0),  # m_y_c_rd is negative
            (500000.0, 0.5, 300.0, 0.0, 355.0, 1.0, 100000000.0),  # t_w is zero
            (500000.0, 0.5, 300.0, 10.0, 355.0, 0.0, 100000000.0),  # gamma_m0 is zero
        ],
    )
    def test_raise_error_when_negative_values_are_given(
        self, w_pl_y: float, rho: float, h_w: float, t_w: float, f_y: float, gamma_m0: float, m_y_c_rd: float
    ) -> None:
        """Test invalid negative values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot30ReducedPlasticResistanceMoment(
                w_pl_y=w_pl_y,
                rho=rho,
                h_w=h_w,
                t_w=t_w,
                f_y=f_y,
                gamma_m0=gamma_m0,
                m_y_c_rd=m_y_c_rd,
            )

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"M_{y,V,Rd} = \min\left(\frac{\left[W_{pl,y} - \frac{\rho \cdot (h_w \cdot t_w)^2}"
                r"{4 \cdot t_w}\right] \cdot f_y}{\gamma_{M0}}, M_{y,c,Rd}\right) = "
                r"\min\left(\frac{\left[500000.000 - \frac{0.500 \cdot (300.000 \cdot 10.000)^2}"
                r"{4 \cdot 10.000}\right] \cdot 355.000}{1.000}, 10000000000.000\right) = 137562500.000 \ Nmm",
            ),
            (
                "complete_with_units",
                r"M_{y,V,Rd} = \min\left(\frac{\left[W_{pl,y} - \frac{\rho \cdot (h_w \cdot t_w)^2}{4 \cdot t_w}\right] "
                r"\cdot f_y}{\gamma_{M0}}, M_{y,c,Rd}\right) = "
                r"\min\left(\frac{\left[500000.000 \ mm^3 - \frac{0.500 \cdot (300.000 \ mm \cdot 10.000 \ mm)^2}"
                r"{4 \cdot 10.000 \ mm}\right] \cdot 355.000 \ MPa}{1.000}, 10000000000.000 \ Nmm\right) = 137562500.000 \ Nmm",
            ),
            ("short", r"M_{y,V,Rd} = 137562500.000 \ Nmm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        w_pl_y = 500000.0  # mm^3
        rho = 0.5  # dimensionless
        h_w = 300.0  # mm
        t_w = 10.0  # mm
        f_y = 355.0  # MPa
        gamma_m0 = 1.0  # dimensionless
        m_y_c_rd = 1e10  # Nmm

        # Object to test
        latex = Form6Dot30ReducedPlasticResistanceMoment(
            w_pl_y=w_pl_y,
            rho=rho,
            h_w=h_w,
            t_w=t_w,
            f_y=f_y,
            gamma_m0=gamma_m0,
            m_y_c_rd=m_y_c_rd,
        ).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Fixtures for testing reinforced concrete sections."""

import pytest

from blueprints.materials.concrete import ConcreteMaterial, ConcreteStrengthClass
from blueprints.materials.reinforcement_steel import ReinforcementSteelMaterial, ReinforcementSteelQuality
from blueprints.structural_sections.concrete.covers import CoversRectangular
from blueprints.structural_sections.concrete.rebar import Rebar
from blueprints.structural_sections.concrete.reinforced_concrete_sections.circular import CircularReinforcedCrossSection
from blueprints.structural_sections.concrete.reinforced_concrete_sections.rectangular import RectangularReinforcedCrossSection
from blueprints.structural_sections.concrete.reinforced_concrete_sections.reinforcement_configurations import (
    ReinforcementByDistance,
    ReinforcementByQuantity,
)


@pytest.fixture
def rectangular_reinforced_cross_section() -> RectangularReinforcedCrossSection:
    """Return a rectangular reinforced cross-section."""
    # Define a concrete material
    concrete = ConcreteMaterial(concrete_class=ConcreteStrengthClass.C35_45)

    # Define a reinforcement steel material
    steel = ReinforcementSteelMaterial(steel_quality=ReinforcementSteelQuality.B500B)

    # Define a rectangular reinforced cross-section
    cs = RectangularReinforcedCrossSection(
        name="Rectangular",
        width=1000,
        height=800,
        covers=CoversRectangular(upper=45, right=30, lower=35, left=50),
        concrete_material=concrete,
    )

    # Add reinforcement to the cross-section
    cs.add_longitudinal_reinforcement_by_quantity(
        n=5,
        diameter=14,
        edge="upper",
        material=steel,
    )
    cs.add_longitudinal_reinforcement_by_quantity(
        n=4,
        diameter=40,
        edge="lower",
        material=steel,
    )

    # Add stirrups to the cross-section
    cs.add_stirrup_along_edges(
        diameter=8,
        distance=150,
        material=steel,
    )
    # Add stirrups to the cross-section
    cs.add_stirrup_along_edges(
        diameter=12,
        distance=300,
        material=steel,
    )

    # Add a longitudinal rebar to the cross-section
    cs.add_longitudinal_rebar(
        rebar=Rebar(
            diameter=12,
            x=-250,
            y=-100,
            material=steel,
        )
    )

    return cs


@pytest.fixture
def circular_reinforced_cross_section() -> CircularReinforcedCrossSection:
    """Return a circular reinforced cross-section."""
    # Define a concrete material
    concrete = ConcreteMaterial(concrete_class=ConcreteStrengthClass.C35_45)

    # Define the reinforcement steel material
    reinforcement_steel = ReinforcementSteelMaterial(steel_quality=ReinforcementSteelQuality.B500B)

    # Define the cross-section
    cs = CircularReinforcedCrossSection(diameter=400, concrete_material=concrete, cover=25)

    # Add longitudinal reinforcement
    cs.add_longitudinal_reinforcement_by_quantity(n=3, diameter=25, material=reinforcement_steel)

    # Add longitudinal reinforcement
    cs.add_longitudinal_reinforcement_by_quantity(n=3, diameter=10, material=reinforcement_steel, start_angle=45)

    # Add stirrups
    cs.add_stirrup_along_perimeter(diameter=25, distance=200, material=reinforcement_steel)

    # Add stirrups
    cs.add_stirrup_along_perimeter(diameter=10, distance=200, material=reinforcement_steel)

    # Add longitudinal rebar in center
    cs.add_longitudinal_rebar(rebar=Rebar(diameter=16, x=0, y=0, material=reinforcement_steel))

    return cs


@pytest.fixture
def reinforcement_by_distance() -> ReinforcementByDistance:
    """Creates a reinforcement by distance configuration."""
    return ReinforcementByDistance(
        diameter=12,
        center_to_center=100,
        material=ReinforcementSteelMaterial(),
    )


@pytest.fixture
def reinforcement_by_quantity() -> ReinforcementByQuantity:
    """Creates a reinforcement by quantity configuration."""
    return ReinforcementByQuantity(
        diameter=12,
        material=ReinforcementSteelMaterial(),
        n=10,
    )



"""Formula 6.27 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM, N_MM, N
from blueprints.validations import raise_if_negative


class Form6Dot27ShearForceInWall(Formula):
    r"""Class representing formula 6.27 for the calculation of the shear force in a wall, [$V_{Ed,i}$]."""

    label = "6.27"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        tau_t_i_t_ef_i: N_MM,
        z_i: MM,
    ) -> None:
        r"""[$V_{Ed,i}$] Shear force in a wall [$N$].

        EN 1992-1-1:2004 art.6.3.2(1) - Formula (6.27)

        Parameters
        ----------
        tau_t_i_t_ef_i : N_MM
            [$\tau_{t,i} t_{ef,i}$] Shear stress in a wall of a section subject to a pure torsional moment multiplied with the
            effective thickness [$N/mm$].
        z_i : MM
            [$z_{i}$] is the side length of wall i defined by the distance between the intersection points with the adjacent walls [$mm$].
        """
        super().__init__()
        self.tau_t_i_t_ef_i = tau_t_i_t_ef_i
        self.z_i = z_i

    @staticmethod
    def _evaluate(
        tau_t_i_t_ef_i: N_MM,
        z_i: MM,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(tau_t_i_t_ef_i=tau_t_i_t_ef_i, z_i=z_i)

        return tau_t_i_t_ef_i * z_i

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.27."""
        return LatexFormula(
            return_symbol=r"V_{Ed,i}",
            result=f"{self:.{n}f}",
            equation=r"\tau_{t,i} t_{ef,i} \cdot z_{i}",
            numeric_equation=rf"{self.tau_t_i_t_ef_i:.{n}f} \cdot {self.z_i:.{n}f}",
            comparison_operator_label="=",
            unit="N",
        )



"""Testing sub-formula 2 for 8.8N of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_8n import (
    SubForm8Dot8nConcreteStress,
    SubForm8Dot8nFunctionY,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestSubForm8Dot8nConcreteStress:
    """Validation for sub-formula 8.8N from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        f_ctd = 2.6  # MPa
        sigma_cm = 15  # MPa
        y_function = 0.5  # -
        f_cd = 25  # MPa
        sub_form_8_8n_2 = SubForm8Dot8nConcreteStress(
            f_ctd=f_ctd,
            sigma_cm=sigma_cm,
            y_function=y_function,
            f_cd=f_cd,
        )
        # Expected result, manually calculated
        manually_result = 35.2
        assert sub_form_8_8n_2 == pytest.approx(expected=manually_result, rel=1e-4)

    def test_evaluation_upper_limit(self) -> None:
        """Test the evaluation of the result if the upper limit is reached."""
        f_ctd = 2.6  # MPa
        sigma_cm = 15  # MPa
        y_function = 0.05  # -
        f_cd = 20  # MPa
        sub_form_8_8n_2 = SubForm8Dot8nConcreteStress(
            f_ctd=f_ctd,
            sigma_cm=sigma_cm,
            y_function=y_function,
            f_cd=f_cd,
        )
        # Expected result, manually calculated
        manually_result = 60
        assert sub_form_8_8n_2 == pytest.approx(expected=manually_result, rel=1e-4)

    def test_raise_error_when_f_ctd_is_negative(self) -> None:
        """Test if a NegativeValueError is raised when f_ctd is negative."""
        # Example values
        f_ctd = -2.6  # MPa
        sigma_cm = 15  # MPa
        y_function = 0.5  # -
        f_cd = 25  # MPa

        with pytest.raises(NegativeValueError):
            SubForm8Dot8nConcreteStress(
                f_ctd=f_ctd,
                sigma_cm=sigma_cm,
                y_function=y_function,
                f_cd=f_cd,
            )

    def test_raise_error_when_sigma_cm_is_negative(self) -> None:
        """Test if a NegativeValueError is raised when sigma_cm is negative."""
        # Example values
        f_ctd = 2.6  # MPa
        sigma_cm = -15  # MPa
        y_function = 0.5  # -
        f_cd = 25  # MPa

        with pytest.raises(NegativeValueError):
            SubForm8Dot8nConcreteStress(
                f_ctd=f_ctd,
                sigma_cm=sigma_cm,
                y_function=y_function,
                f_cd=f_cd,
            )

    def test_raise_error_when_y_is_negative(self) -> None:
        """Test if a LessOrEqualToZeroError is raised when y is negative."""
        # Example values
        f_ctd = 2.6  # MPa
        sigma_cm = 15  # MPa
        y_function = -0.5  # -
        f_cd = 25  # MPa

        with pytest.raises(LessOrEqualToZeroError):
            SubForm8Dot8nConcreteStress(
                f_ctd=f_ctd,
                sigma_cm=sigma_cm,
                y_function=y_function,
                f_cd=f_cd,
            )

    def test_raise_error_when_y_is_zero(self) -> None:
        """Test if a LessOrEqualToZeroError is raised when y is zero."""
        # Example values
        f_ctd = 2.6  # MPa
        sigma_cm = 15  # MPa
        y_function = 0  # -
        f_cd = 25  # MPa

        with pytest.raises(LessOrEqualToZeroError):
            SubForm8Dot8nConcreteStress(
                f_ctd=f_ctd,
                sigma_cm=sigma_cm,
                y_function=y_function,
                f_cd=f_cd,
            )

    def test_raise_error_when_f_cd_is_negative(self) -> None:
        """Test if a NegativeValueError is raised when f_cd is negative."""
        # Example values
        f_ctd = 2.6  # MPa
        sigma_cm = 15  # MPa
        y_function = 0.5  # -
        f_cd = -25  # MPa

        with pytest.raises(NegativeValueError):
            SubForm8Dot8nConcreteStress(
                f_ctd=f_ctd,
                sigma_cm=sigma_cm,
                y_function=y_function,
                f_cd=f_cd,
            )

    def test_integration_with_sub_form_8_8n_function_y(self) -> None:
        """Test the integration with sub-formula 8.8 for calculating function y."""
        # Example values
        f_ctd = 2.6  # MPa
        sigma_cm = 15  # MPa
        f_cd = 50  # MPa
        x_function = 1.5  # -
        y_function = SubForm8Dot8nFunctionY(x_function=x_function)

        # Object to test
        sub_form_8_8n_2 = SubForm8Dot8nConcreteStress(
            f_ctd=f_ctd,
            sigma_cm=sigma_cm,
            y_function=y_function,
            f_cd=f_cd,
        )

        # Expected result, manually calculated
        manually_result = 144.412473

        assert sub_form_8_8n_2 == pytest.approx(expected=manually_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                (
                    r"\sigma_{td} = \min\left( 3 \cdot f_{cd}, \frac{f_{ctd} + \sigma_{cm}}{y} \right)"
                    r" = \min\left(3 \cdot 50.00, \frac{2.60 + 15.00}{0.12} \right) = 144.41"
                ),
            ),
            ("short", r"\sigma_{td} = 144.41"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ctd = 2.6  # MPa
        sigma_cm = 15  # MPa
        f_cd = 50  # MPa
        x_function = 1.5  # -
        y_function = SubForm8Dot8nFunctionY(x_function=x_function)

        # Object to test
        sub_form_8_8n_2_latex = SubForm8Dot8nConcreteStress(
            f_ctd=f_ctd,
            sigma_cm=sigma_cm,
            y_function=y_function,
            f_cd=f_cd,
        ).latex()

        actual = {"complete": sub_form_8_8n_2_latex.complete, "short": sub_form_8_8n_2_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Base class of all steel cross-sections."""

from abc import ABC
from collections.abc import Sequence
from dataclasses import dataclass, field

from shapely.geometry import Polygon
from shapely.geometry.base import BaseGeometry
from shapely.geometry.polygon import orient

from blueprints.structural_sections._cross_section import CrossSection
from blueprints.structural_sections.steel.steel_element import SteelElement
from blueprints.type_alias import KG_M, M3_M, MM, MPA
from blueprints.unit_conversion import MM3_TO_M3


@dataclass(kw_only=True)
class CombinedSteelCrossSection(CrossSection, ABC):
    """Base class of all steel cross-sections.

    Parameters
    ----------
    elements : Sequence[SteelElement], optional
        A sequence of steel elements that make up the cross-section.
        Default is an empty list.
    name : str, optional
        The name of the cross-section. Default is "Combined Steel Cross Section".
    """

    elements: Sequence[SteelElement] = field(default_factory=list)
    name: str = "Combined Steel Cross Section"

    @property
    def polygon(self) -> Polygon:
        """Return the polygon of the steel cross-section."""
        # check if there are any elements
        if not self.elements:
            raise ValueError("No elements have been added to the cross-section.")

        # return the polygon of the first element if there is only one
        if len(self.elements) == 1:
            return self.elements[0].cross_section.polygon

        # Combine the polygons of all elements if there is multiple
        combined_polygon: BaseGeometry = self.elements[0].cross_section.polygon
        for element in self.elements[1:]:
            combined_polygon = combined_polygon.union(element.cross_section.polygon)

        # Ensure the result is a valid Polygon
        if not isinstance(combined_polygon, Polygon):
            raise TypeError("The combined geometry is not a valid Polygon.")

        # Ensure consistent orientation
        return orient(combined_polygon)

    @property
    def height(self) -> MM:
        """Height of the cross-section [mm]."""
        return self.polygon.bounds[3] - self.polygon.bounds[1]

    @property
    def width(self) -> MM:
        """Width of the cross-section [mm]."""
        return self.polygon.bounds[2] - self.polygon.bounds[0]

    @property
    def volume_per_meter(self) -> M3_M:
        """Total volume of the reinforced cross-section per meter length [m³/m]."""
        length = 1000  # mm
        return self.area * length * MM3_TO_M3

    @property
    def weight_per_meter(self) -> KG_M:
        """
        Calculate the weight per meter of the steel element.

        Returns
        -------
        KG_M
            The weight per meter of the steel element.
        """
        return sum(element.weight_per_meter for element in self.elements)

    @property
    def yield_strength(self) -> MPA:
        """
        Calculate the yield strength of the steel element.

        This is the minimum yield strength of all elements in the cross-section.

        Returns
        -------
        MPa
            The yield strength of the steel element.
        """
        return min(element.yield_strength for element in self.elements)

    @property
    def ultimate_strength(self) -> MPA:
        """
        Calculate the ultimate strength of the steel element.

        This is the minimum ultimate strength of all elements in the cross-section.

        Returns
        -------
        MPa
            The ultimate strength of the steel element.
        """
        return min(element.ultimate_strength for element in self.elements)



"""Testing formula 6.72 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_72 import (
    Form6Dot72FatigueResistanceConcreteCompression,
)
from blueprints.validations import NegativeValueError


class TestForm6Dot72FatigueResistanceConcreteCompression:
    """Validation for formula 6.72 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        e_cd_max_equ = 0.8
        r_equ = 0.5

        # Object to test
        formula = Form6Dot72FatigueResistanceConcreteCompression(e_cd_max_equ=e_cd_max_equ, r_equ=r_equ)

        # Expected result, manually calculated
        expected_result = False

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("e_cd_max_equ", "r_equ"),
        [
            (-0.8, 0.5),  # e_cd_max_equ is negative
            (0.8, -0.5),  # r_equ is negative
            (0.8, 1.5),  # r_equ is greater than 1
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, e_cd_max_equ: float, r_equ: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot72FatigueResistanceConcreteCompression(e_cd_max_equ=e_cd_max_equ, r_equ=r_equ)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to E_{cd,max,equ} + 0.43 \cdot \sqrt{1 - R_{equ}} \leq 1 \to "
                r"0.800 + 0.43 \cdot \sqrt{1 - 0.500} \leq 1 \to \text{Not OK}",
            ),
            ("short", r"CHECK \to \text{Not OK}"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        e_cd_max_equ = 0.8
        r_equ = 0.5

        # Object to test
        latex = Form6Dot72FatigueResistanceConcreteCompression(e_cd_max_equ=e_cd_max_equ, r_equ=r_equ).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.51 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_51 import (
    Form6Dot51AppliedPunchingShearStressEccentricLoading,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot51AppliedPunchingShearStressEccentricLoading:
    """Validation for formula 6.51 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        v_ed_red = 10000.0
        u = 200.0
        d = 300.0
        k = 1.5
        m_ed = 400.0
        w = 500.0

        # Object to test
        formula = Form6Dot51AppliedPunchingShearStressEccentricLoading(v_ed_red=v_ed_red, u=u, d=d, k=k, m_ed=m_ed, w=w)

        # Expected result, manually calculated
        manually_calculated_result = 0.17066666666  # MPa

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("v_ed_red", "u", "d", "k", "m_ed", "w"),
        [
            (-10000.0, 200.0, 300.0, 1.5, 400.0, 500.0),  # v_ed_red is negative
            (10000.0, -200.0, 300.0, 1.5, 400.0, 500.0),  # u is negative
            (10000.0, 200.0, -300.0, 1.5, 400.0, 500.0),  # d is negative
            (10000.0, 200.0, 300.0, -1.5, 400.0, 500.0),  # k is negative
            (10000.0, 200.0, 300.0, 1.5, -400.0, 500.0),  # m_ed is negative
            (10000.0, 200.0, 300.0, 1.5, 400.0, -500.0),  # w is negative
            (0.0, 200.0, 300.0, 1.5, 400.0, 500.0),  # v_ed_red is zero
            (10000.0, 0.0, 300.0, 1.5, 400.0, 500.0),  # u is zero
            (10000.0, 200.0, 0.0, 1.5, 400.0, 500.0),  # d is zero
            (10000.0, 200.0, 300.0, 1.5, 400.0, 0.0),  # w is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, v_ed_red: float, u: float, d: float, k: float, m_ed: float, w: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot51AppliedPunchingShearStressEccentricLoading(v_ed_red=v_ed_red, u=u, d=d, k=k, m_ed=m_ed, w=w)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"v_{Ed} = \frac{V_{Ed,red}}{u \cdot d} \cdot \left(1 + k \cdot \frac{M_{Ed} \cdot u}{V_{Ed,red} \cdot W}\right) = "
                r"\frac{10000.000}{200.000 \cdot 300.000} \cdot \left(1 + 1.500 \cdot \frac{400.000 \cdot 200.000}{10000.000 \cdot 500.000}"
                r"\right) = 0.171 \ MPa",
            ),
            ("short", r"v_{Ed} = 0.171 \ MPa"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        v_ed_red = 10000.0
        u = 200.0
        d = 300.0
        k = 1.5
        m_ed = 400.0
        w = 500.0

        # Object to test
        latex = Form6Dot51AppliedPunchingShearStressEccentricLoading(v_ed_red=v_ed_red, u=u, d=d, k=k, m_ed=m_ed, w=w).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 8.14 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_14 import (
    Form8Dot14EquivalentDiameterBundledBars,
)
from blueprints.validations import NegativeValueError


class TestForm8Dot14EquivalentDiameterBundledBars:
    """Validation for formula 8.14 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        diameter = 16  # mm
        n_b = 4  # [-]
        form_8_14 = Form8Dot14EquivalentDiameterBundledBars(diameter=diameter, n_b=n_b)

        # manually calculated result
        manually_calculated_result = 32  # mm

        assert form_8_14 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_maximum(self) -> None:
        """Test the evaluation of the result if the maximum is reached."""
        # example values
        diameter = 32  # mm
        n_b = 4  # [-]
        form_8_14 = Form8Dot14EquivalentDiameterBundledBars(diameter=diameter, n_b=n_b)

        # manually calculated result
        manually_calculated_result = 55  # mm

        assert form_8_14 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_if_diameter_is_negative(self) -> None:
        """Test that a NegativeValueError is raised if diameter is negative."""
        # example values
        diameter = -32  # mm
        n_b = 4  # [-]

        with pytest.raises(NegativeValueError):
            Form8Dot14EquivalentDiameterBundledBars(diameter=diameter, n_b=n_b)

    def test_raise_error_if_n_b_is_negative(self) -> None:
        """Test that a NegativeValueError is raised if n_b is negative."""
        # example values
        diameter = 32
        n_b = -4

        with pytest.raises(NegativeValueError):
            Form8Dot14EquivalentDiameterBundledBars(diameter=diameter, n_b=n_b)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                (r"Ø_n = \min \left(55 \ \text{mm}, Ø \cdot \sqrt{n_b} \right) = \min \left(55 \ \text{mm}, 16.00 \cdot \sqrt{4.00} \right) = 32.00"),
            ),
            ("short", r"Ø_n = 32.00"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        diameter = 16  # mm
        n_b = 4  # [-]

        # Object to test
        form_8_14_latex = Form8Dot14EquivalentDiameterBundledBars(diameter=diameter, n_b=n_b).latex()

        actual = {"complete": form_8_14_latex.complete, "short": form_8_14_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 8.17 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_16 import (
    Form8Dot16BasicTransmissionLength,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_17 import (
    Form8Dot17DesignValueTransmissionLength1,
)
from blueprints.validations import NegativeValueError


class TestForm8Dot17DesignValueTransmissionLength1:
    """Validation for formula 8.17 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        l_pt = 120  # mm
        form_8_17 = Form8Dot17DesignValueTransmissionLength1(l_pt=l_pt)
        # manually calculated result
        manually_calculated_result = 96  # mm

        assert form_8_17 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_if_negative_l_pt(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for l_pt."""
        l_pt = -1
        with pytest.raises(NegativeValueError):
            Form8Dot17DesignValueTransmissionLength1(l_pt=l_pt)

    def test_integration_with_form_8_dot_16(self) -> None:
        """Test integration between formula 8.16 and 8.17."""
        alpha_1 = 1  # [-]
        alpha_2 = 0.25  # [-]
        diameter = 8  # mm
        sigma_pm0 = 350  # MPa
        f_bpt = 5  # MPa
        l_pt = Form8Dot16BasicTransmissionLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            diameter=diameter,
            sigma_pm0=sigma_pm0,
            f_bpt=f_bpt,
        )

        form_8_17 = Form8Dot17DesignValueTransmissionLength1(l_pt=l_pt)
        manually_calculated_result = 112  # mm

        assert form_8_17 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            ("complete", r"l_{pt1} = 0.8 \cdot l_{pt} = 0.8 \cdot 140.000 = 112.000"),
            ("short", r"l_{pt1} = 112.000"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 0.25  # [-]
        diameter = 8  # mm
        sigma_pm0 = 350  # MPa
        f_bpt = 5  # MPa
        l_pt = Form8Dot16BasicTransmissionLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            diameter=diameter,
            sigma_pm0=sigma_pm0,
            f_bpt=f_bpt,
        )

        # Object to test
        form_8_17_latex = Form8Dot17DesignValueTransmissionLength1(l_pt=l_pt).latex()

        actual = {"complete": form_8_17_latex.complete, "short": form_8_17_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."
